# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The model definition for Continuous 2D layers

Adapted from: https://github.com/CompVis/stable-diffusion/blob/
21f890f9da3cfbeaba8e2ac3c425ee9e998d5229/ldm/modules/diffusionmodules/model.py

[Copyright (c) 2022 Robin Rombach and Patrick Esser and contributors]
https://github.com/CompVis/stable-diffusion/blob/
21f890f9da3cfbeaba8e2ac3c425ee9e998d5229/LICENSE
"""

import math

import numpy as np

# pytorch_diffusion + derived encoder decoder
import torch
import torch.nn as nn
import torch.nn.functional as F

from uniception.models.libs.cosmos_tokenizer.modules.patching import Patcher, UnPatcher
from uniception.models.libs.cosmos_tokenizer.modules.utils import Normalize, nonlinearity


class Upsample(nn.Module):
    def __init__(self, in_channels: int):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, stride=1, padding=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.repeat_interleave(2, dim=2).repeat_interleave(2, dim=3)
        return self.conv(x)


class Downsample(nn.Module):
    def __init__(self, in_channels: int):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, stride=2, padding=0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pad = (0, 1, 0, 1)
        x = F.pad(x, pad, mode="constant", value=0)
        return self.conv(x)


class ResnetBlock(nn.Module):
    def __init__(
        self,
        *,
        in_channels: int,
        out_channels: int = None,
        dropout: float,
        **kwargs,
    ):
        super().__init__()
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels

        self.norm1 = Normalize(in_channels)
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=1)
        self.norm2 = Normalize(out_channels)
        self.dropout = nn.Dropout(dropout)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1)
        self.nin_shortcut = (
            nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0)
            if in_channels != out_channels
            else nn.Identity()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        h = x
        h = self.norm1(h)
        h = nonlinearity(h)
        h = self.conv1(h)

        h = self.norm2(h)
        h = nonlinearity(h)
        h = self.dropout(h)
        h = self.conv2(h)

        x = self.nin_shortcut(x)

        return x + h


class AttnBlock(nn.Module):
    def __init__(self, in_channels: int):
        super().__init__()

        self.norm = Normalize(in_channels)
        self.q = nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)
        self.k = nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)
        self.v = nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)
        self.proj_out = nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # TODO (freda): Consider reusing implementations in Attn `imaginaire`,
        # since than one is gonna be based on TransformerEngine's attn op,
        # w/c could ease CP implementations.
        h_ = x
        h_ = self.norm(h_)
        q = self.q(h_)
        k = self.k(h_)
        v = self.v(h_)

        # compute attention
        b, c, h, w = q.shape
        q = q.reshape(b, c, h * w)
        q = q.permute(0, 2, 1)
        k = k.reshape(b, c, h * w)
        w_ = torch.bmm(q, k)
        w_ = w_ * (int(c) ** (-0.5))
        w_ = F.softmax(w_, dim=2)

        # attend to values
        v = v.reshape(b, c, h * w)
        w_ = w_.permute(0, 2, 1)
        h_ = torch.bmm(v, w_)
        h_ = h_.reshape(b, c, h, w)

        h_ = self.proj_out(h_)

        return x + h_


class Encoder(nn.Module):
    def __init__(
        self,
        in_channels: int,
        channels: int,
        channels_mult: list[int],
        num_res_blocks: int,
        attn_resolutions: list[int],
        dropout: float,
        resolution: int,
        z_channels: int,
        spatial_compression: int,
        **ignore_kwargs,
    ):
        super().__init__()
        self.num_resolutions = len(channels_mult)
        self.num_res_blocks = num_res_blocks

        # Patcher.
        patch_size = ignore_kwargs.get("patch_size", 1)
        self.patcher = Patcher(patch_size, ignore_kwargs.get("patch_method", "rearrange"))
        in_channels = in_channels * patch_size * patch_size

        # calculate the number of downsample operations
        self.num_downsamples = int(math.log2(spatial_compression)) - int(math.log2(patch_size))
        assert (
            self.num_downsamples <= self.num_resolutions
        ), f"we can only downsample {self.num_resolutions} times at most"

        # downsampling
        self.conv_in = torch.nn.Conv2d(in_channels, channels, kernel_size=3, stride=1, padding=1)

        curr_res = resolution // patch_size
        in_ch_mult = (1,) + tuple(channels_mult)
        self.in_ch_mult = in_ch_mult
        self.down = nn.ModuleList()
        for i_level in range(self.num_resolutions):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_in = channels * in_ch_mult[i_level]
            block_out = channels * channels_mult[i_level]
            for _ in range(self.num_res_blocks):
                block.append(
                    ResnetBlock(
                        in_channels=block_in,
                        out_channels=block_out,
                        dropout=dropout,
                    )
                )
                block_in = block_out
                if curr_res in attn_resolutions:
                    attn.append(AttnBlock(block_in))
            down = nn.Module()
            down.block = block
            down.attn = attn
            if i_level < self.num_downsamples:
                down.downsample = Downsample(block_in)
                curr_res = curr_res // 2
            self.down.append(down)

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(in_channels=block_in, out_channels=block_in, dropout=dropout)
        self.mid.attn_1 = AttnBlock(block_in)
        self.mid.block_2 = ResnetBlock(in_channels=block_in, out_channels=block_in, dropout=dropout)

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = torch.nn.Conv2d(block_in, z_channels, kernel_size=3, stride=1, padding=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patcher(x)

        # downsampling
        hs = [self.conv_in(x)]
        for i_level in range(self.num_resolutions):
            for i_block in range(self.num_res_blocks):
                h = self.down[i_level].block[i_block](hs[-1])
                if len(self.down[i_level].attn) > 0:
                    h = self.down[i_level].attn[i_block](h)
                hs.append(h)
            if i_level < self.num_downsamples:
                hs.append(self.down[i_level].downsample(hs[-1]))

        # middle
        h = hs[-1]
        h = self.mid.block_1(h)
        h = self.mid.attn_1(h)
        h = self.mid.block_2(h)

        # end
        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        return h


class Decoder(nn.Module):
    def __init__(
        self,
        out_channels: int,
        channels: int,
        channels_mult: list[int],
        num_res_blocks: int,
        attn_resolutions: int,
        dropout: float,
        resolution: int,
        z_channels: int,
        spatial_compression: int,
        **ignore_kwargs,
    ):
        super().__init__()
        self.num_resolutions = len(channels_mult)
        self.num_res_blocks = num_res_blocks

        # UnPatcher.
        patch_size = ignore_kwargs.get("patch_size", 1)
        self.unpatcher = UnPatcher(patch_size, ignore_kwargs.get("patch_method", "rearrange"))
        out_ch = out_channels * patch_size * patch_size

        # calculate the number of upsample operations
        self.num_upsamples = int(math.log2(spatial_compression)) - int(math.log2(patch_size))
        assert self.num_upsamples <= self.num_resolutions, f"we can only upsample {self.num_resolutions} times at most"

        block_in = channels * channels_mult[self.num_resolutions - 1]
        curr_res = (resolution // patch_size) // 2 ** (self.num_resolutions - 1)
        self.z_shape = (1, z_channels, curr_res, curr_res)

        # z to block_in
        self.conv_in = torch.nn.Conv2d(z_channels, block_in, kernel_size=3, stride=1, padding=1)

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(in_channels=block_in, out_channels=block_in, dropout=dropout)
        self.mid.attn_1 = AttnBlock(block_in)
        self.mid.block_2 = ResnetBlock(in_channels=block_in, out_channels=block_in, dropout=dropout)

        # upsampling
        self.up = nn.ModuleList()
        for i_level in reversed(range(self.num_resolutions)):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_out = channels * channels_mult[i_level]
            for _ in range(self.num_res_blocks + 1):
                block.append(
                    ResnetBlock(
                        in_channels=block_in,
                        out_channels=block_out,
                        dropout=dropout,
                    )
                )
                block_in = block_out
                if curr_res in attn_resolutions:
                    attn.append(AttnBlock(block_in))
            up = nn.Module()
            up.block = block
            up.attn = attn
            if i_level >= (self.num_resolutions - self.num_upsamples):
                up.upsample = Upsample(block_in)
                curr_res = curr_res * 2
            self.up.insert(0, up)

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = torch.nn.Conv2d(block_in, out_ch, kernel_size=3, stride=1, padding=1)

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        h = self.conv_in(z)

        # middle
        h = self.mid.block_1(h)
        h = self.mid.attn_1(h)
        h = self.mid.block_2(h)

        # upsampling
        for i_level in reversed(range(self.num_resolutions)):
            for i_block in range(self.num_res_blocks + 1):
                h = self.up[i_level].block[i_block](h)
                if len(self.up[i_level].attn) > 0:
                    h = self.up[i_level].attn[i_block](h)
            if i_level >= (self.num_resolutions - self.num_upsamples):
                h = self.up[i_level].upsample(h)

        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        h = self.unpatcher(h)
        return h



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The patcher and unpatcher implementation for 2D and 3D data.

The idea of Haar wavelet is to compute LL, LH, HL, HH component as two 1D convolutions.
One on the rows and one on the columns.
For example, in 1D signal, we have [a, b], then the low-freq compoenent is [a + b] / 2 and high-freq is [a - b] / 2.
We can use a 1D convolution with kernel [1, 1] and stride 2 to represent the L component.
For H component, we can use a 1D convolution with kernel [1, -1] and stride 2.
Although in principle, we typically only do additional Haar wavelet over the LL component. But here we do it for all
   as we need to support downsampling for more than 2x.
For example, 4x downsampling can be done by 2x Haar and additional 2x Haar, and the shape would be.
   [3, 256, 256] -> [12, 128, 128] -> [48, 64, 64]
"""

import torch
import torch.nn.functional as F
from einops import rearrange

_WAVELETS = {
    "haar": torch.tensor([0.7071067811865476, 0.7071067811865476]),
    "rearrange": torch.tensor([1.0, 1.0]),
}
_PERSISTENT = False


class Patcher(torch.nn.Module):
    """A module to convert image tensors into patches using torch operations.

    The main difference from `class Patching` is that this module implements
    all operations using torch, rather than python or numpy, for efficiency purpose.

    It's bit-wise identical to the Patching module outputs, with the added
    benefit of being torch.jit scriptable.
    """

    def __init__(self, patch_size=1, patch_method="haar"):
        super().__init__()
        self.patch_size = patch_size
        self.patch_method = patch_method
        self.register_buffer("wavelets", _WAVELETS[patch_method], persistent=_PERSISTENT)
        self.range = range(int(torch.log2(torch.tensor(self.patch_size)).item()))
        self.register_buffer(
            "_arange",
            torch.arange(_WAVELETS[patch_method].shape[0]),
            persistent=_PERSISTENT,
        )
        for param in self.parameters():
            param.requires_grad = False

    def forward(self, x):
        if self.patch_method == "haar":
            return self._haar(x)
        elif self.patch_method == "rearrange":
            return self._arrange(x)
        else:
            raise ValueError("Unknown patch method: " + self.patch_method)

    def _dwt(self, x, mode="reflect", rescale=False):
        dtype = x.dtype
        h = self.wavelets

        n = h.shape[0]
        g = x.shape[1]
        hl = h.flip(0).reshape(1, 1, -1).repeat(g, 1, 1)
        hh = (h * ((-1) ** self._arange)).reshape(1, 1, -1).repeat(g, 1, 1)
        hh = hh.to(dtype=dtype)
        hl = hl.to(dtype=dtype)

        x = F.pad(x, pad=(n - 2, n - 1, n - 2, n - 1), mode=mode).to(dtype)
        xl = F.conv2d(x, hl.unsqueeze(2), groups=g, stride=(1, 2))
        xh = F.conv2d(x, hh.unsqueeze(2), groups=g, stride=(1, 2))
        xll = F.conv2d(xl, hl.unsqueeze(3), groups=g, stride=(2, 1))
        xlh = F.conv2d(xl, hh.unsqueeze(3), groups=g, stride=(2, 1))
        xhl = F.conv2d(xh, hl.unsqueeze(3), groups=g, stride=(2, 1))
        xhh = F.conv2d(xh, hh.unsqueeze(3), groups=g, stride=(2, 1))

        out = torch.cat([xll, xlh, xhl, xhh], dim=1)
        if rescale:
            out = out / 2
        return out

    def _haar(self, x):
        for _ in self.range:
            x = self._dwt(x, rescale=True)
        return x

    def _arrange(self, x):
        x = rearrange(
            x,
            "b c (h p1) (w p2) -> b (c p1 p2) h w",
            p1=self.patch_size,
            p2=self.patch_size,
        ).contiguous()
        return x


class Patcher3D(Patcher):
    """A 3D discrete wavelet transform for video data, expects 5D tensor, i.e. a batch of videos."""

    def __init__(self, patch_size=1, patch_method="haar"):
        super().__init__(patch_method=patch_method, patch_size=patch_size)
        self.register_buffer(
            "patch_size_buffer",
            patch_size * torch.ones([1], dtype=torch.int32),
            persistent=_PERSISTENT,
        )

    def _dwt(self, x, wavelet, mode="reflect", rescale=False):
        dtype = x.dtype
        h = self.wavelets

        n = h.shape[0]
        g = x.shape[1]
        hl = h.flip(0).reshape(1, 1, -1).repeat(g, 1, 1)
        hh = (h * ((-1) ** self._arange)).reshape(1, 1, -1).repeat(g, 1, 1)
        hh = hh.to(dtype=dtype)
        hl = hl.to(dtype=dtype)

        # Handles temporal axis.
        x = F.pad(x, pad=(max(0, n - 2), n - 1, n - 2, n - 1, n - 2, n - 1), mode=mode).to(dtype)
        xl = F.conv3d(x, hl.unsqueeze(3).unsqueeze(4), groups=g, stride=(2, 1, 1))
        xh = F.conv3d(x, hh.unsqueeze(3).unsqueeze(4), groups=g, stride=(2, 1, 1))

        # Handles spatial axes.
        xll = F.conv3d(xl, hl.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))
        xlh = F.conv3d(xl, hh.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))
        xhl = F.conv3d(xh, hl.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))
        xhh = F.conv3d(xh, hh.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))

        xlll = F.conv3d(xll, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xllh = F.conv3d(xll, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xlhl = F.conv3d(xlh, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xlhh = F.conv3d(xlh, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xhll = F.conv3d(xhl, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xhlh = F.conv3d(xhl, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xhhl = F.conv3d(xhh, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xhhh = F.conv3d(xhh, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))

        out = torch.cat([xlll, xllh, xlhl, xlhh, xhll, xhlh, xhhl, xhhh], dim=1)
        if rescale:
            out = out / (2 * torch.sqrt(torch.tensor(2.0)))
        return out

    def _haar(self, x):
        xi, xv = torch.split(x, [1, x.shape[2] - 1], dim=2)
        x = torch.cat([xi.repeat_interleave(self.patch_size, dim=2), xv], dim=2)
        for _ in self.range:
            x = self._dwt(x, "haar", rescale=True)
        return x

    def _arrange(self, x):
        xi, xv = torch.split(x, [1, x.shape[2] - 1], dim=2)
        x = torch.cat([xi.repeat_interleave(self.patch_size, dim=2), xv], dim=2)
        x = rearrange(
            x,
            "b c (t p1) (h p2) (w p3) -> b (c p1 p2 p3) t h w",
            p1=self.patch_size,
            p2=self.patch_size,
            p3=self.patch_size,
        ).contiguous()
        return x


class UnPatcher(torch.nn.Module):
    """A module to convert patches into image tensorsusing torch operations.

    The main difference from `class Unpatching` is that this module implements
    all operations using torch, rather than python or numpy, for efficiency purpose.

    It's bit-wise identical to the Unpatching module outputs, with the added
    benefit of being torch.jit scriptable.
    """

    def __init__(self, patch_size=1, patch_method="haar"):
        super().__init__()
        self.patch_size = patch_size
        self.patch_method = patch_method
        self.register_buffer("wavelets", _WAVELETS[patch_method], persistent=_PERSISTENT)
        self.range = range(int(torch.log2(torch.tensor(self.patch_size)).item()))
        self.register_buffer(
            "_arange",
            torch.arange(_WAVELETS[patch_method].shape[0]),
            persistent=_PERSISTENT,
        )
        for param in self.parameters():
            param.requires_grad = False

    def forward(self, x):
        if self.patch_method == "haar":
            return self._ihaar(x)
        elif self.patch_method == "rearrange":
            return self._iarrange(x)
        else:
            raise ValueError("Unknown patch method: " + self.patch_method)

    def _idwt(self, x, wavelet="haar", mode="reflect", rescale=False):
        dtype = x.dtype
        h = self.wavelets
        n = h.shape[0]

        g = x.shape[1] // 4
        hl = h.flip([0]).reshape(1, 1, -1).repeat([g, 1, 1])
        hh = (h * ((-1) ** self._arange)).reshape(1, 1, -1).repeat(g, 1, 1)
        hh = hh.to(dtype=dtype)
        hl = hl.to(dtype=dtype)

        xll, xlh, xhl, xhh = torch.chunk(x.to(dtype), 4, dim=1)

        # Inverse transform.
        yl = torch.nn.functional.conv_transpose2d(xll, hl.unsqueeze(3), groups=g, stride=(2, 1), padding=(n - 2, 0))
        yl += torch.nn.functional.conv_transpose2d(xlh, hh.unsqueeze(3), groups=g, stride=(2, 1), padding=(n - 2, 0))
        yh = torch.nn.functional.conv_transpose2d(xhl, hl.unsqueeze(3), groups=g, stride=(2, 1), padding=(n - 2, 0))
        yh += torch.nn.functional.conv_transpose2d(xhh, hh.unsqueeze(3), groups=g, stride=(2, 1), padding=(n - 2, 0))
        y = torch.nn.functional.conv_transpose2d(yl, hl.unsqueeze(2), groups=g, stride=(1, 2), padding=(0, n - 2))
        y += torch.nn.functional.conv_transpose2d(yh, hh.unsqueeze(2), groups=g, stride=(1, 2), padding=(0, n - 2))

        if rescale:
            y = y * 2
        return y

    def _ihaar(self, x):
        for _ in self.range:
            x = self._idwt(x, "haar", rescale=True)
        return x

    def _iarrange(self, x):
        x = rearrange(
            x,
            "b (c p1 p2) h w -> b c (h p1) (w p2)",
            p1=self.patch_size,
            p2=self.patch_size,
        )
        return x


class UnPatcher3D(UnPatcher):
    """A 3D inverse discrete wavelet transform for video wavelet decompositions."""

    def __init__(self, patch_size=1, patch_method="haar"):
        super().__init__(patch_method=patch_method, patch_size=patch_size)

    def _idwt(self, x, wavelet="haar", mode="reflect", rescale=False):
        dtype = x.dtype
        h = self.wavelets
        n = h.shape[0]

        g = x.shape[1] // 8  # split into 8 spatio-temporal filtered tesnors.
        hl = h.flip([0]).reshape(1, 1, -1).repeat([g, 1, 1])
        hh = (h * ((-1) ** self._arange)).reshape(1, 1, -1).repeat(g, 1, 1)
        hl = hl.to(dtype=dtype)
        hh = hh.to(dtype=dtype)

        xlll, xllh, xlhl, xlhh, xhll, xhlh, xhhl, xhhh = torch.chunk(x, 8, dim=1)

        # Height height transposed convolutions.
        xll = F.conv_transpose3d(xlll, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xll += F.conv_transpose3d(xllh, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))

        xlh = F.conv_transpose3d(xlhl, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xlh += F.conv_transpose3d(xlhh, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))

        xhl = F.conv_transpose3d(xhll, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xhl += F.conv_transpose3d(xhlh, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))

        xhh = F.conv_transpose3d(xhhl, hl.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))
        xhh += F.conv_transpose3d(xhhh, hh.unsqueeze(2).unsqueeze(3), groups=g, stride=(1, 1, 2))

        # Handles width transposed convolutions.
        xl = F.conv_transpose3d(xll, hl.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))
        xl += F.conv_transpose3d(xlh, hh.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))
        xh = F.conv_transpose3d(xhl, hl.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))
        xh += F.conv_transpose3d(xhh, hh.unsqueeze(2).unsqueeze(4), groups=g, stride=(1, 2, 1))

        # Handles time axis transposed convolutions.
        x = F.conv_transpose3d(xl, hl.unsqueeze(3).unsqueeze(4), groups=g, stride=(2, 1, 1))
        x += F.conv_transpose3d(xh, hh.unsqueeze(3).unsqueeze(4), groups=g, stride=(2, 1, 1))

        if rescale:
            x = x * (2 * torch.sqrt(torch.tensor(2.0)))
        return x

    def _ihaar(self, x):
        for _ in self.range:
            x = self._idwt(x, "haar", rescale=True)
        x = x[:, :, self.patch_size - 1 :, ...]
        return x

    def _iarrange(self, x):
        x = rearrange(
            x,
            "b (c p1 p2 p3) t h w -> b c (t p1) (h p2) (w p3)",
            p1=self.patch_size,
            p2=self.patch_size,
            p3=self.patch_size,
        )
        x = x[:, :, self.patch_size - 1 :, ...]
        return x



# Copyright (C) 2022-present Naver Corporation. All rights reserved.
# Licensed under CC BY-NC-SA 4.0 (non-commercial use only).

from setuptools import setup
from torch import cuda
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

# compile for all possible CUDA architectures
all_cuda_archs = cuda.get_gencode_flags().replace("compute=", "arch=").split()
# alternatively, you can list cuda archs that you want, eg:
# all_cuda_archs = [
# '-gencode', 'arch=compute_70,code=sm_70',
# '-gencode', 'arch=compute_75,code=sm_75',
# '-gencode', 'arch=compute_80,code=sm_80',
# '-gencode', 'arch=compute_86,code=sm_86'
# ]

setup(
    name="curope",
    ext_modules=[
        CUDAExtension(
            name="curope",
            sources=[
                "curope.cpp",
                "kernels.cu",
            ],
            extra_compile_args=dict(
                nvcc=["-O3", "--ptxas-options=-v", "--use_fast_math"] + all_cuda_archs, cxx=["-O3"]
            ),
        )
    ],
    cmdclass={"build_ext": BuildExtension},
)



"""
UniCeption Alternating-Attention Transformer for Information Sharing
"""

from functools import partial
from typing import Callable, List, Optional, Tuple, Type, Union

import numpy as np
import torch
import torch.nn as nn

from uniception.models.info_sharing.base import (
    MultiViewTransformerInput,
    MultiViewTransformerOutput,
    UniCeptionInfoSharingBase,
)
from uniception.models.utils.intermediate_feature_return import IntermediateFeatureReturner, feature_take_indices
from uniception.models.utils.positional_encoding import PositionGetter
from uniception.models.utils.transformer_blocks import Mlp, SelfAttentionBlock


class MultiViewAlternatingAttentionTransformer(UniCeptionInfoSharingBase):
    "UniCeption Multi-View Alternating-Attention Transformer for information sharing across image features from different views."

    def __init__(
        self,
        name: str,
        input_embed_dim: int,
        use_pe_for_non_reference_views: bool = False,
        max_num_views_for_pe: int = 1000,
        use_rand_idx_pe_for_non_reference_views: bool = True,
        size: Optional[str] = None,
        depth: int = 12,
        dim: int = 768,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_norm: bool = False,
        proj_drop: float = 0.0,
        attn_drop: float = 0.0,
        init_values: Optional[float] = None,
        drop_path: float = 0.0,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Union[Type[nn.Module], Callable[..., nn.Module]] = partial(nn.LayerNorm, eps=1e-6),
        mlp_layer: Type[nn.Module] = Mlp,
        custom_positional_encoding: Optional[Callable] = None,
        use_scalable_softmax: bool = False,
        use_entropy_scaling: bool = False,
        base_token_count_for_entropy_scaling: int = 444,
        entropy_scaling_growth_factor: float = 1.4,
        pretrained_checkpoint_path: Optional[str] = None,
        gradient_checkpointing: bool = False,
        *args,
        **kwargs,
    ):
        """
        Initialize the Multi-View Alternating-Attention Transformer for information sharing across image features from different views.
        Alternates between global and frame-level attention.

        Args:
            input_embed_dim (int): Dimension of input embeddings.
            use_pe_for_non_reference_views (bool): Whether to use view positional encoding for input non-reference views. (default: False)
            max_num_views_for_pe (int): Maximum number of views for positional encoding. (default: 1000)
            use_rand_idx_pe_for_non_reference_views (bool): Whether to use random index positional encoding for non-reference views. (default: True)
            size (str): String to indicate interpretable size of the transformer (for e.g., base, large, ...). (default: None)
            depth (int): Number of transformer layers. (default: 12, base size)
            dim (int): Dimension of the transformer. (default: 768, base size)
            num_heads (int): Number of attention heads. (default: 12, base size)
            mlp_ratio (float): Ratio of hidden to input dimension in MLP (default: 4.)
            qkv_bias (bool): Whether to include bias in qkv projection (default: True)
            qk_norm (bool): Whether to normalize q and k (default: False)
            proj_drop (float): Dropout rate for output (default: 0.)
            attn_drop (float): Dropout rate for attention weights (default: 0.)
            init_values (float): Initial value for LayerScale gamma (default: None)
            drop_path (float): Dropout rate for stochastic depth (default: 0.)
            act_layer (nn.Module): Activation layer (default: nn.GELU)
            norm_layer (nn.Module): Normalization layer (default: nn.LayerNorm)
            mlp_layer (nn.Module): MLP layer (default: Mlp)
            custom_positional_encoding (Callable): Custom positional encoding function (default: None)
            use_scalable_softmax (bool): Whether to use scalable softmax (default: False)
            use_entropy_scaling (bool): Whether to use entropy scaling (default: False)
            base_token_count_for_entropy_scaling (int): Base token count for entropy scaling (default: 444)
                                                        Computed using (518, 168) as base resolution with 14 patch size
            entropy_scaling_growth_factor (float): Growth factor for entropy scaling (default: 1.4)
            pretrained_checkpoint_path (str, optional): Path to the pretrained checkpoint. (default: None)
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing for memory efficiency. (default: False)
        """
        # Initialize the base class
        super().__init__(name=name, size=size, *args, **kwargs)

        # Initialize the specific attributes of the transformer
        self.input_embed_dim = input_embed_dim
        self.use_pe_for_non_reference_views = use_pe_for_non_reference_views
        self.max_num_views_for_pe = max_num_views_for_pe
        self.use_rand_idx_pe_for_non_reference_views = use_rand_idx_pe_for_non_reference_views
        self.depth = depth
        self.dim = dim
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.qk_norm = qk_norm
        self.proj_drop = proj_drop
        self.attn_drop = attn_drop
        self.init_values = init_values
        self.drop_path = drop_path
        self.act_layer = act_layer
        self.norm_layer = norm_layer
        self.mlp_layer = mlp_layer
        self.custom_positional_encoding = custom_positional_encoding
        self.use_scalable_softmax = use_scalable_softmax
        self.use_entropy_scaling = use_entropy_scaling
        self.base_token_count_for_entropy_scaling = base_token_count_for_entropy_scaling
        self.entropy_scaling_growth_factor = entropy_scaling_growth_factor
        self.pretrained_checkpoint_path = pretrained_checkpoint_path
        self.gradient_checkpointing = gradient_checkpointing

        # Initialize the projection layer for input embeddings
        if self.input_embed_dim != self.dim:
            self.proj_embed = nn.Linear(self.input_embed_dim, self.dim, bias=True)
        else:
            self.proj_embed = nn.Identity()

        # Initialize the self-attention blocks which ingest all views at once
        self.self_attention_blocks = nn.ModuleList(
            [
                SelfAttentionBlock(
                    dim=self.dim,
                    num_heads=self.num_heads,
                    mlp_ratio=self.mlp_ratio,
                    qkv_bias=self.qkv_bias,
                    qk_norm=self.qk_norm,
                    proj_drop=self.proj_drop,
                    attn_drop=self.attn_drop,
                    init_values=self.init_values,
                    drop_path=self.drop_path,
                    act_layer=self.act_layer,
                    norm_layer=self.norm_layer,
                    mlp_layer=self.mlp_layer,
                    custom_positional_encoding=self.custom_positional_encoding,
                    use_scalable_softmax=self.use_scalable_softmax,
                    use_entropy_scaling=self.use_entropy_scaling,
                    base_token_count_for_entropy_scaling=self.base_token_count_for_entropy_scaling,
                    entropy_scaling_growth_factor=self.entropy_scaling_growth_factor,
                )
                for _ in range(self.depth)
            ]
        )

        # Initialize the final normalization layer
        self.norm = self.norm_layer(self.dim)

        # Initialize the position getter for patch positions if required
        if self.custom_positional_encoding is not None:
            self.position_getter = PositionGetter()

        if self.use_pe_for_non_reference_views:
            # Initialize the positional encoding table for the different views
            self.register_buffer(
                "view_pos_table",
                self._get_sinusoid_encoding_table(self.max_num_views_for_pe, self.dim, 10000),
            )
        else:
            # Initialize the positional encoding table for the reference view
            self.register_buffer(
                "view_pos_table",
                self._get_sinusoid_encoding_table(1, self.dim, 10000),
            )

        # Initialize random weights
        self.initialize_weights()

        # Apply gradient checkpointing if enabled
        if self.gradient_checkpointing:
            for i, block in enumerate(self.self_attention_blocks):
                self.self_attention_blocks[i] = self.wrap_module_with_gradient_checkpointing(block)

        # Load pretrained weights if provided
        if self.pretrained_checkpoint_path is not None:
            print(
                f"Loading pretrained multi-view Alternating-Attention transformer weights from {self.pretrained_checkpoint_path} ..."
            )
            ckpt = torch.load(self.pretrained_checkpoint_path, weights_only=False)
            print(self.load_state_dict(ckpt["model"]))

    def _get_sinusoid_encoding_table(self, n_position, d_hid, base):
        "Sinusoid position encoding table"

        def get_position_angle_vec(position):
            return [position / np.power(base, 2 * (hid_j // 2) / d_hid) for hid_j in range(d_hid)]

        sinusoid_table = np.array([get_position_angle_vec(pos_i) for pos_i in range(n_position)])
        sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])
        sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])

        return torch.FloatTensor(sinusoid_table)

    def initialize_weights(self):
        "Initialize weights of the transformer."
        # Linears and layer norms
        self.apply(self._init_weights)

    def _init_weights(self, m):
        "Initialize the transformer linear and layer norm weights."
        if isinstance(m, nn.Linear):
            # We use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(
        self,
        model_input: MultiViewTransformerInput,
    ) -> MultiViewTransformerOutput:
        """
        Forward interface for the Multi-View Alternating-Attention Transformer.

        Args:
            model_input (MultiViewTransformerInput): Input to the model.
                Expects the features to be a list of size (batch, input_embed_dim, height, width),
                where each entry corresponds to a different view.
                Optionally, the input can also include additional_input_tokens (e.g., class token, registers, pose tokens, scale token)
                which are appended to the token set from the multi-view features. The tokens are of size (batch, input_embed_dim, num_of_additional_tokens).

        Returns:
            MultiViewTransformerOutput: Output of the model post information sharing.
        """
        # Check that the number of views matches the input and the features are of expected shape
        if self.use_pe_for_non_reference_views:
            assert (
                len(model_input.features) <= self.max_num_views_for_pe
            ), f"Expected less than {self.max_num_views_for_pe} views, got {len(model_input.features)}"
        assert all(
            view_features.shape[1] == self.input_embed_dim for view_features in model_input.features
        ), f"All views must have input dimension {self.input_embed_dim}"
        assert all(
            view_features.ndim == 4 for view_features in model_input.features
        ), "All views must have 4 dimensions (N, C, H, W)"

        # Initialize the multi-view features from the model input and number of views for current input
        multi_view_features = model_input.features
        num_of_views = len(multi_view_features)
        batch_size, _, height, width = multi_view_features[0].shape
        num_of_tokens_per_view = height * width

        # Stack the multi-view features (N, C, H, W) to (N, V, C, H, W) (assumes all V views have same shape)
        multi_view_features = torch.stack(multi_view_features, dim=1)

        # Resize the multi-view features from NVCHW to NLC, where L = V * H * W
        multi_view_features = multi_view_features.permute(0, 1, 3, 4, 2)  # (N, V, H, W, C)
        multi_view_features = multi_view_features.reshape(
            batch_size, num_of_views * height * width, self.input_embed_dim
        ).contiguous()

        # Process additional input tokens if provided
        if model_input.additional_input_tokens is not None:

            additional_tokens = model_input.additional_input_tokens
            assert additional_tokens.ndim == 3, "Additional tokens must have 3 dimensions (N, C, T)"
            assert (
                additional_tokens.shape[1] == self.input_embed_dim
            ), f"Additional tokens must have input dimension {self.input_embed_dim}"
            assert additional_tokens.shape[0] == batch_size, "Batch size mismatch for additional tokens"

            # Reshape to channel-last format for transformer processing
            additional_tokens = additional_tokens.permute(0, 2, 1).contiguous()  # (N, C, T) -> (N, T, C)

            # Concatenate the additional tokens to the multi-view features
            multi_view_features = torch.cat([multi_view_features, additional_tokens], dim=1)

        # Project input features to the transformer dimension
        multi_view_features = self.proj_embed(multi_view_features)

        # Create patch positions for each view if custom positional encoding is used
        if self.custom_positional_encoding is not None:
            multi_view_positions = [
                self.position_getter(batch_size, height, width, multi_view_features.device)
            ] * num_of_views  # List of length V, where each tensor is (N, H * W, C)
            multi_view_positions = torch.cat(multi_view_positions, dim=1)  # (N, V * H * W, C)
        else:
            multi_view_positions = [None] * num_of_views

        # Add None positions for additional tokens if they exist
        if model_input.additional_input_tokens is not None:

            additional_tokens_positions = [None] * model_input.additional_input_tokens.shape[1]
            multi_view_positions = multi_view_positions + additional_tokens_positions

        # Add positional encoding for reference view (idx 0)
        ref_view_pe = self.view_pos_table[0].clone().detach()
        ref_view_pe = ref_view_pe.reshape((1, 1, self.dim))
        ref_view_pe = ref_view_pe.repeat(batch_size, num_of_tokens_per_view, 1)
        ref_view_features = multi_view_features[:, :num_of_tokens_per_view, :]
        ref_view_features = ref_view_features + ref_view_pe

        if self.use_pe_for_non_reference_views:
            # Add positional encoding for non-reference views (sequential indices starting from idx 1 or random indices which are uniformly sampled)
            if self.use_rand_idx_pe_for_non_reference_views:
                non_ref_view_pe_indices = torch.randint(low=1, high=self.max_num_views_for_pe, size=(num_of_views - 1,))
            else:
                non_ref_view_pe_indices = torch.arange(1, num_of_views)
            non_ref_view_pe = self.view_pos_table[non_ref_view_pe_indices].clone().detach()
            non_ref_view_pe = non_ref_view_pe.reshape((1, num_of_views - 1, self.dim))
            non_ref_view_pe = non_ref_view_pe.repeat_interleave(num_of_tokens_per_view, dim=1)
            non_ref_view_pe = non_ref_view_pe.repeat(batch_size, 1, 1)
            non_ref_view_features = multi_view_features[
                :, num_of_tokens_per_view : num_of_views * num_of_tokens_per_view, :
            ]
            non_ref_view_features = non_ref_view_features + non_ref_view_pe
        else:
            non_ref_view_features = multi_view_features[
                :, num_of_tokens_per_view : num_of_views * num_of_tokens_per_view, :
            ]

        # Concatenate the reference and non-reference view features
        # Handle additional tokens (no view-based positional encoding for them)
        if model_input.additional_input_tokens is not None:

            additional_features = multi_view_features[:, num_of_views * num_of_tokens_per_view :, :]
            multi_view_features = torch.cat([ref_view_features, non_ref_view_features, additional_features], dim=1)
        else:
            multi_view_features = torch.cat([ref_view_features, non_ref_view_features], dim=1)

        # Loop over the depth of the transformer
        for depth_idx in range(self.depth):
            if depth_idx % 2 == 0:
                # Apply the self-attention block and update the multi-view features
                # Global attention across all views
                multi_view_features = self.self_attention_blocks[depth_idx](multi_view_features, multi_view_positions)
            else:
                # Handle additional tokens separately for frame-level attention
                additional_features = None
                additional_positions = None
                if model_input.additional_input_tokens is not None:

                    # Extract additional token features
                    additional_features = multi_view_features[:, num_of_views * num_of_tokens_per_view :, :]
                    # Keep only view features for frame-level attention
                    multi_view_features = multi_view_features[:, : num_of_views * num_of_tokens_per_view, :]

                    # Handle positions for additional tokens if custom positional encoding is used
                    if self.custom_positional_encoding is not None:
                        additional_positions = multi_view_positions[:, num_of_views * num_of_tokens_per_view :, :]
                        multi_view_positions = multi_view_positions[:, : num_of_views * num_of_tokens_per_view, :]

                # Reshape the multi-view features from (N, V * H * W, C) to (N * V, H * W, C)
                multi_view_features = multi_view_features.reshape(
                    batch_size * num_of_views, num_of_tokens_per_view, self.dim
                ).contiguous()  # (N * V, H * W, C)
                if multi_view_positions[0] is not None:
                    multi_view_positions = multi_view_positions.reshape(
                        batch_size * num_of_views, num_of_tokens_per_view, 2
                    ).contiguous()  # (N * V, H * W, C)

                # Apply the self-attention block and update the multi-view features
                # Frame-level attention within each view
                multi_view_features = self.self_attention_blocks[depth_idx](multi_view_features, multi_view_positions)

                # Reshape the multi-view features from (N * V, H * W, C) back to (N, V * H * W, C)
                multi_view_features = multi_view_features.reshape(
                    batch_size, num_of_views * num_of_tokens_per_view, self.dim
                ).contiguous()  # (N, V * H * W, C)
                if multi_view_positions[0] is not None:
                    multi_view_positions = multi_view_positions.reshape(
                        batch_size, num_of_views * num_of_tokens_per_view, 2
                    ).contiguous()  # (N, V * H * W, C)

                # Reattach additional tokens if they exist
                if additional_features is not None:
                    multi_view_features = torch.cat([multi_view_features, additional_features], dim=1)
                    # Reattach positions for additional tokens if they exist
                    if additional_positions is not None:
                        multi_view_positions = torch.cat([multi_view_positions, additional_positions], dim=1)

        # Normalize the output features
        output_multi_view_features = self.norm(multi_view_features)

        # Extract only the view features (excluding additional tokens)
        view_features = output_multi_view_features[:, : num_of_views * num_of_tokens_per_view, :]

        # Reshape the output multi-view features (N, V * H * W, C) back to (N, V, C, H, W)
        view_features = view_features.reshape(batch_size, num_of_views, height, width, self.dim)  # (N, V, H, W, C)
        view_features = view_features.permute(0, 1, 4, 2, 3).contiguous()  # (N, V, C, H, W)

        # Split the output multi-view features into separate views
        view_features = view_features.split(1, dim=1)
        view_features = [output_view_features.squeeze(dim=1) for output_view_features in view_features]

        # Extract and return additional token features if provided
        if model_input.additional_input_tokens is not None:

            additional_token_features = output_multi_view_features[:, num_of_views * num_of_tokens_per_view :, :]
            additional_token_features = additional_token_features.permute(0, 2, 1).contiguous()  # (N, C, T)
            return MultiViewTransformerOutput(
                features=view_features, additional_token_features=additional_token_features
            )
        else:
            return MultiViewTransformerOutput(features=view_features)


class MultiViewAlternatingAttentionTransformerIFR(
    MultiViewAlternatingAttentionTransformer, IntermediateFeatureReturner
):
    "Intermediate Feature Returner for UniCeption Multi-View Alternating-Attention Transformer"

    def __init__(
        self,
        name: str,
        input_embed_dim: int,
        use_pe_for_non_reference_views: bool = False,
        max_num_views_for_pe: int = 1000,
        use_rand_idx_pe_for_non_reference_views: bool = True,
        size: Optional[str] = None,
        depth: int = 12,
        dim: int = 768,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_norm: bool = False,
        proj_drop: float = 0.0,
        attn_drop: float = 0.0,
        init_values: Optional[float] = None,
        drop_path: float = 0.0,
        act_layer: nn.Module = nn.GELU,
        norm_layer: nn.Module = partial(nn.LayerNorm, eps=1e-6),
        mlp_layer: nn.Module = Mlp,
        custom_positional_encoding: Callable = None,
        use_scalable_softmax: bool = False,
        use_entropy_scaling: bool = False,
        base_token_count_for_entropy_scaling: int = 444,
        entropy_scaling_growth_factor: float = 1.4,
        pretrained_checkpoint_path: str = None,
        indices: Optional[Union[int, List[int]]] = None,
        norm_intermediate: bool = True,
        intermediates_only: bool = False,
        gradient_checkpointing: bool = False,
        *args,
        **kwargs,
    ):
        """
        Initialize the Multi-View Alternating-Attention Transformer for information sharing across image features from different views.
        Extends the base class to return intermediate features.

        Args:
            input_embed_dim (int): Dimension of input embeddings.
            use_pe_for_non_reference_views (bool): Whether to use view positional encoding for input non-reference views. (default: False)
            max_num_views_for_pe (int): Maximum number of views for positional encoding. (default: 1000)
            use_rand_idx_pe_for_non_reference_views (bool): Whether to use random index positional encoding for non-reference views. (default: True)
            size (str): String to indicate interpretable size of the transformer (for e.g., base, large, ...). (default: None)
            depth (int): Number of transformer layers. (default: 12, base size)
            dim (int): Dimension of the transformer. (default: 768, base size)
            num_heads (int): Number of attention heads. (default: 12, base size)
            mlp_ratio (float): Ratio of hidden to input dimension in MLP (default: 4.)
            qkv_bias (bool): Whether to include bias in qkv projection (default: False)
            qk_norm (bool): Whether to normalize q and k (default: False)
            proj_drop (float): Dropout rate for output (default: 0.)
            attn_drop (float): Dropout rate for attention weights (default: 0.)
            init_values (float): Initial value for LayerScale gamma (default: None)
            drop_path (float): Dropout rate for stochastic depth (default: 0.)
            act_layer (nn.Module): Activation layer (default: nn.GELU)
            norm_layer (nn.Module): Normalization layer (default: nn.LayerNorm)
            mlp_layer (nn.Module): MLP layer (default: Mlp)
            custom_positional_encoding (Callable): Custom positional encoding function (default: None)
            use_scalable_softmax (bool): Whether to use scalable softmax (default: False)
            use_entropy_scaling (bool): Whether to use entropy scaling (default: False)
            base_token_count_for_entropy_scaling (int): Base token count for entropy scaling (default: 444)
                                                        Computed using (518, 168) as base resolution with 14 patch size
            entropy_scaling_growth_factor (float): Growth factor for entropy scaling (default: 1.4)
            pretrained_checkpoint_path (str, optional): Path to the pretrained checkpoint. (default: None)
            indices (Optional[Union[int, List[int]]], optional): Indices of the layers to return. (default: None) Options:
            - None: Return all intermediate layers.
            - int: Return the last n layers.
            - List[int]: Return the intermediate layers at the specified indices.
            norm_intermediate (bool, optional): Whether to normalize the intermediate features. (default: True)
            intermediates_only (bool, optional): Whether to return only the intermediate features. (default: False)
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing for memory efficiency. (default: False)
        """
        # Init the base classes
        MultiViewAlternatingAttentionTransformer.__init__(
            self,
            name=name,
            input_embed_dim=input_embed_dim,
            use_pe_for_non_reference_views=use_pe_for_non_reference_views,
            max_num_views_for_pe=max_num_views_for_pe,
            use_rand_idx_pe_for_non_reference_views=use_rand_idx_pe_for_non_reference_views,
            size=size,
            depth=depth,
            dim=dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            proj_drop=proj_drop,
            attn_drop=attn_drop,
            init_values=init_values,
            drop_path=drop_path,
            act_layer=act_layer,
            norm_layer=norm_layer,
            mlp_layer=mlp_layer,
            custom_positional_encoding=custom_positional_encoding,
            use_scalable_softmax=use_scalable_softmax,
            use_entropy_scaling=use_entropy_scaling,
            base_token_count_for_entropy_scaling=base_token_count_for_entropy_scaling,
            entropy_scaling_growth_factor=entropy_scaling_growth_factor,
            pretrained_checkpoint_path=pretrained_checkpoint_path,
            gradient_checkpointing=gradient_checkpointing,
            *args,
            **kwargs,
        )
        IntermediateFeatureReturner.__init__(
            self,
            indices=indices,
            norm_intermediate=norm_intermediate,
            intermediates_only=intermediates_only,
        )

    def forward(
        self,
        model_input: MultiViewTransformerInput,
    ) -> Union[
        List[MultiViewTransformerOutput],
        Tuple[MultiViewTransformerOutput, List[MultiViewTransformerOutput]],
    ]:
        """
        Forward interface for the Multi-View Alternating-Attention Transformer with Intermediate Feature Return.

        Args:
            model_input (MultiViewTransformerInput): Input to the model.
                Expects the features to be a list of size (batch, input_embed_dim, height, width),
                where each entry corresponds to a different view.
                Optionally, the input can also include additional_input_tokens (e.g., class token, registers, pose tokens, scale token)
                which are appended to the token set from the multi-view features. The tokens are of size (batch, input_embed_dim, num_of_additional_tokens).

        Returns:
            Union[List[MultiViewTransformerOutput], Tuple[MultiViewTransformerOutput, List[MultiViewTransformerOutput]]]:
                Output of the model post information sharing.
                If intermediates_only is True, returns a list of intermediate outputs.
                If intermediates_only is False, returns a tuple of final output and a list of intermediate outputs.
        """
        # Check that the number of views matches the input and the features are of expected shape
        if self.use_pe_for_non_reference_views:
            assert (
                len(model_input.features) <= self.max_num_views_for_pe
            ), f"Expected less than {self.max_num_views_for_pe} views, got {len(model_input.features)}"
        assert all(
            view_features.shape[1] == self.input_embed_dim for view_features in model_input.features
        ), f"All views must have input dimension {self.input_embed_dim}"
        assert all(
            view_features.ndim == 4 for view_features in model_input.features
        ), "All views must have 4 dimensions (N, C, H, W)"

        # Get the indices of the intermediate features to return
        intermediate_multi_view_features = []
        take_indices, _ = feature_take_indices(self.depth, self.indices)

        # Initialize the multi-view features from the model input and number of views for current input
        multi_view_features = model_input.features
        num_of_views = len(multi_view_features)
        batch_size, _, height, width = multi_view_features[0].shape
        num_of_tokens_per_view = height * width

        # Stack the multi-view features (N, C, H, W) to (N, V, C, H, W) (assumes all V views have same shape)
        multi_view_features = torch.stack(multi_view_features, dim=1)

        # Resize the multi-view features from NVCHW to NLC, where L = V * H * W
        multi_view_features = multi_view_features.permute(0, 1, 3, 4, 2)  # (N, V, H, W, C)
        multi_view_features = multi_view_features.reshape(
            batch_size, num_of_views * height * width, self.input_embed_dim
        ).contiguous()

        # Process additional input tokens if provided
        if model_input.additional_input_tokens is not None:

            additional_tokens = model_input.additional_input_tokens
            assert additional_tokens.ndim == 3, "Additional tokens must have 3 dimensions (N, C, T)"
            assert (
                additional_tokens.shape[1] == self.input_embed_dim
            ), f"Additional tokens must have input dimension {self.input_embed_dim}"
            assert additional_tokens.shape[0] == batch_size, "Batch size mismatch for additional tokens"

            # Reshape to channel-last format for transformer processing
            additional_tokens = additional_tokens.permute(0, 2, 1).contiguous()  # (N, C, T) -> (N, T, C)

            # Concatenate the additional tokens to the multi-view features
            multi_view_features = torch.cat([multi_view_features, additional_tokens], dim=1)

        # Project input features to the transformer dimension
        multi_view_features = self.proj_embed(multi_view_features)

        # Create patch positions for each view if custom positional encoding is used
        if self.custom_positional_encoding is not None:
            multi_view_positions = [
                self.position_getter(batch_size, height, width, multi_view_features.device)
            ] * num_of_views  # List of length V, where each tensor is (N, H * W, C)
            multi_view_positions = torch.cat(multi_view_positions, dim=1)  # (N, V * H * W, C)
        else:
            multi_view_positions = [None] * num_of_views

        # Add None positions for additional tokens if they exist
        if model_input.additional_input_tokens is not None:

            additional_tokens_positions = [None] * model_input.additional_input_tokens.shape[1]
            multi_view_positions = multi_view_positions + additional_tokens_positions

        # Add positional encoding for reference view (idx 0)
        ref_view_pe = self.view_pos_table[0].clone().detach()
        ref_view_pe = ref_view_pe.reshape((1, 1, self.dim))
        ref_view_pe = ref_view_pe.repeat(batch_size, num_of_tokens_per_view, 1)
        ref_view_features = multi_view_features[:, :num_of_tokens_per_view, :]
        ref_view_features = ref_view_features + ref_view_pe

        if self.use_pe_for_non_reference_views:
            # Add positional encoding for non-reference views (sequential indices starting from idx 1 or random indices which are uniformly sampled)
            if self.use_rand_idx_pe_for_non_reference_views:
                non_ref_view_pe_indices = torch.randint(low=1, high=self.max_num_views_for_pe, size=(num_of_views - 1,))
            else:
                non_ref_view_pe_indices = torch.arange(1, num_of_views)
            non_ref_view_pe = self.view_pos_table[non_ref_view_pe_indices].clone().detach()
            non_ref_view_pe = non_ref_view_pe.reshape((1, num_of_views - 1, self.dim))
            non_ref_view_pe = non_ref_view_pe.repeat_interleave(num_of_tokens_per_view, dim=1)
            non_ref_view_pe = non_ref_view_pe.repeat(batch_size, 1, 1)
            non_ref_view_features = multi_view_features[
                :, num_of_tokens_per_view : num_of_views * num_of_tokens_per_view, :
            ]
            non_ref_view_features = non_ref_view_features + non_ref_view_pe
        else:
            non_ref_view_features = multi_view_features[
                :, num_of_tokens_per_view : num_of_views * num_of_tokens_per_view, :
            ]

        # Concatenate the reference and non-reference view features
        # Handle additional tokens (no view-based positional encoding for them)
        if model_input.additional_input_tokens is not None:

            additional_features = multi_view_features[:, num_of_views * num_of_tokens_per_view :, :]
            multi_view_features = torch.cat([ref_view_features, non_ref_view_features, additional_features], dim=1)
        else:
            multi_view_features = torch.cat([ref_view_features, non_ref_view_features], dim=1)

        # Loop over the depth of the transformer
        for depth_idx in range(self.depth):
            if depth_idx % 2 == 0:
                # Apply the self-attention block and update the multi-view features
                # Global attention across all views
                multi_view_features = self.self_attention_blocks[depth_idx](multi_view_features, multi_view_positions)
            else:
                # Handle additional tokens separately for frame-level attention
                additional_features = None
                additional_positions = None
                if model_input.additional_input_tokens is not None:

                    # Extract additional token features
                    additional_features = multi_view_features[:, num_of_views * num_of_tokens_per_view :, :]
                    # Keep only view features for frame-level attention
                    multi_view_features = multi_view_features[:, : num_of_views * num_of_tokens_per_view, :]

                    # Handle positions for additional tokens if custom positional encoding is used
                    if self.custom_positional_encoding is not None:
                        additional_positions = multi_view_positions[:, num_of_views * num_of_tokens_per_view :, :]
                        multi_view_positions = multi_view_positions[:, : num_of_views * num_of_tokens_per_view, :]

                # Reshape the multi-view features from (N, V * H * W, C) to (N * V, H * W, C)
                multi_view_features = multi_view_features.reshape(
                    batch_size * num_of_views, num_of_tokens_per_view, self.dim
                ).contiguous()  # (N * V, H * W, C)
                if multi_view_positions[0] is not None:
                    multi_view_positions = multi_view_positions.reshape(
                        batch_size * num_of_views, num_of_tokens_per_view, 2
                    ).contiguous()  # (N * V, H * W, C)

                # Apply the self-attention block and update the multi-view features
                # Frame-level attention within each view
                multi_view_features = self.self_attention_blocks[depth_idx](multi_view_features, multi_view_positions)

                # Reshape the multi-view features from (N * V, H * W, C) back to (N, V * H * W, C)
                multi_view_features = multi_view_features.reshape(
                    batch_size, num_of_views * num_of_tokens_per_view, self.dim
                ).contiguous()  # (N, V * H * W, C)
                if multi_view_positions[0] is not None:
                    multi_view_positions = multi_view_positions.reshape(
                        batch_size, num_of_views * num_of_tokens_per_view, 2
                    ).contiguous()  # (N, V * H * W, C)

                # Reattach additional tokens if they exist
                if additional_features is not None:
                    multi_view_features = torch.cat([multi_view_features, additional_features], dim=1)
                    # Reattach positions for additional tokens if they exist
                    if additional_positions is not None:
                        multi_view_positions = torch.cat([multi_view_positions, additional_positions], dim=1)
            if depth_idx in take_indices:
                # Normalize the intermediate features with final norm layer if enabled
                intermediate_multi_view_features.append(
                    self.norm(multi_view_features) if self.norm_intermediate else multi_view_features
                )

        # Reshape the intermediate features and convert to MultiViewTransformerOutput class
        for idx in range(len(intermediate_multi_view_features)):
            # Get the current intermediate features
            current_features = intermediate_multi_view_features[idx]

            # Extract additional token features if provided
            additional_token_features = None
            if model_input.additional_input_tokens is not None:

                additional_token_features = current_features[:, num_of_views * num_of_tokens_per_view :, :]
                additional_token_features = additional_token_features.permute(0, 2, 1).contiguous()  # (N, C, T)
                # Only keep the view features for reshaping
                current_features = current_features[:, : num_of_views * num_of_tokens_per_view, :]

            # Reshape the intermediate multi-view features (N, V * H * W, C) back to (N, V, C, H, W)
            current_features = current_features.reshape(
                batch_size, num_of_views, height, width, self.dim
            )  # (N, V, H, W, C)
            current_features = current_features.permute(0, 1, 4, 2, 3).contiguous()  # (N, V, C, H, W)

            # Split the intermediate multi-view features into separate views
            current_features = current_features.split(1, dim=1)
            current_features = [
                intermediate_view_features.squeeze(dim=1) for intermediate_view_features in current_features
            ]

            intermediate_multi_view_features[idx] = MultiViewTransformerOutput(
                features=current_features, additional_token_features=additional_token_features
            )

        # Return only the intermediate features if enabled
        if self.intermediates_only:
            return intermediate_multi_view_features

        # Normalize the output features
        output_multi_view_features = self.norm(multi_view_features)

        # Extract view features (excluding additional tokens)
        additional_token_features = None
        if model_input.additional_input_tokens is not None:

            additional_token_features = output_multi_view_features[:, num_of_views * num_of_tokens_per_view :, :]
            additional_token_features = additional_token_features.permute(0, 2, 1).contiguous()  # (N, C, T)
            view_features = output_multi_view_features[:, : num_of_views * num_of_tokens_per_view, :]
        else:
            view_features = output_multi_view_features

        # Reshape the output multi-view features (N, V * H * W, C) back to (N, V, C, H, W)
        view_features = view_features.reshape(batch_size, num_of_views, height, width, self.dim)  # (N, V, H, W, C)
        view_features = view_features.permute(0, 1, 4, 2, 3).contiguous()  # (N, V, C, H, W)

        # Split the output multi-view features into separate views
        view_features = view_features.split(1, dim=1)
        view_features = [output_view_features.squeeze(dim=1) for output_view_features in view_features]

        output_multi_view_features = MultiViewTransformerOutput(
            features=view_features, additional_token_features=additional_token_features
        )

        return output_multi_view_features, intermediate_multi_view_features


def dummy_positional_encoding(x, xpos):
    "Dummy function for positional encoding of tokens"
    x = x
    xpos = xpos
    return x


def test_reshape_for_frame_attention():
    "Test the reshape function for frame-level attention in the Alternating Attention Transformer"
    batch_size = 2
    num_of_views = 3
    height = width = 2
    dim = 4
    num_of_tokens_per_view = height * width

    # Create tensor with recognizable pattern
    x = torch.zeros(batch_size, num_of_views * num_of_tokens_per_view, dim)
    for b in range(batch_size):
        for v in range(num_of_views):
            for h in range(height):
                for w in range(width):
                    token_idx = v * num_of_tokens_per_view + h * width + w
                    x[b, token_idx] = torch.tensor([b, v, h, w])

    # Apply reshape
    reshaped = x.reshape(batch_size * num_of_views, num_of_tokens_per_view, dim).contiguous()

    # Verify shape
    assert reshaped.shape == (batch_size * num_of_views, num_of_tokens_per_view, dim)

    # Verify content (check a few values)
    for b in range(batch_size):
        for v in range(num_of_views):
            for h in range(height):
                for w in range(width):
                    batch_view_idx = b * num_of_views + v
                    token_idx = h * width + w
                    expected = torch.tensor([b, v, h, w])
                    assert torch.all(reshaped[batch_view_idx, token_idx] == expected)

    # Verify reshape back works
    back_to_original = reshaped.reshape(batch_size, num_of_views * num_of_tokens_per_view, dim)
    assert torch.all(x == back_to_original)

    print("Reshape test passed!")


if __name__ == "__main__":
    # Unit test the reshape logic used for frame-level attention
    test_reshape_for_frame_attention()

    # Init multi-view alternating-attention transformer with no custom positional encoding and run a forward pass
    for num_views in [2, 3, 4]:
        print(f"Testing MultiViewAlternatingAttentionTransformer with {num_views} views ...")
        # No positional encoding for non-reference views
        model = MultiViewAlternatingAttentionTransformer(
            name="MV-AAT",
            input_embed_dim=1024,
        )
        model_input = [torch.rand(1, 1024, 14, 14) for _ in range(num_views)]
        model_input = MultiViewTransformerInput(features=model_input)
        model_output = model(model_input)
        assert len(model_output.features) == num_views
        assert all(f.shape == (1, model.dim, 14, 14) for f in model_output.features)
        # Sequential idx based positional encoding
        model = MultiViewAlternatingAttentionTransformer(
            name="MV-AAT",
            input_embed_dim=1024,
            use_pe_for_non_reference_views=True,
            max_num_views_for_pe=1000,
            use_rand_idx_pe_for_non_reference_views=False,
        )
        model_input = [torch.rand(1, 1024, 14, 14) for _ in range(num_views)]
        model_input = MultiViewTransformerInput(features=model_input)
        model_output = model(model_input)
        assert len(model_output.features) == num_views
        assert all(f.shape == (1, model.dim, 14, 14) for f in model_output.features)
        # Random idx based positional encoding
        model = MultiViewAlternatingAttentionTransformer(
            name="MV-AAT",
            input_embed_dim=1024,
            use_pe_for_non_reference_views=True,
            max_num_views_for_pe=1000,
            use_rand_idx_pe_for_non_reference_views=True,
        )
        model_input = [torch.rand(1, 1024, 14, 14) for _ in range(num_views)]
        model_input = MultiViewTransformerInput(features=model_input)
        model_output = model(model_input)
        assert len(model_output.features) == num_views
        assert all(f.shape == (1, model.dim, 14, 14) for f in model_output.features)

    # Init multi-view alternating-attention transformer with custom positional encoding and run a forward pass
    for num_views in [2, 3, 4]:
        print(
            f"Testing MultiViewAlternatingAttentionTransformer with {num_views} views and custom positional encoding ..."
        )
        model = MultiViewAlternatingAttentionTransformer(
            name="MV-AAT",
            input_embed_dim=1024,
            custom_positional_encoding=dummy_positional_encoding,
        )
        model_input = [torch.rand(1, 1024, 14, 14) for _ in range(num_views)]
        model_input = MultiViewTransformerInput(features=model_input)
        model_output = model(model_input)
        assert len(model_output.features) == num_views
        assert all(f.shape == (1, model.dim, 14, 14) for f in model_output.features)

    print("All multi-view alternating-attention transformers initialized and tested successfully!")

    # Intermediate Feature Returner Tests
    print("Running Intermediate Feature Returner Tests ...")

    # Run the intermediate feature returner with last-n index
    model_intermediate_feature_returner = MultiViewAlternatingAttentionTransformerIFR(
        name="MV-AAT-IFR",
        input_embed_dim=1024,
        indices=6,  # Last 6 layers
    )
    model_input = [torch.rand(1, 1024, 14, 14) for _ in range(2)]
    model_input = MultiViewTransformerInput(features=model_input)
    output = model_intermediate_feature_returner(model_input)
    assert isinstance(output, tuple)
    assert isinstance(output[0], MultiViewTransformerOutput)
    assert len(output[1]) == 6
    assert all(isinstance(intermediate, MultiViewTransformerOutput) for intermediate in output[1])
    assert len(output[1][0].features) == 2

    # Run the intermediate feature returner with specific indices
    model_intermediate_feature_returner = MultiViewAlternatingAttentionTransformerIFR(
        name="MV-AAT-IFR",
        input_embed_dim=1024,
        indices=[0, 2, 4, 6],  # Specific indices
    )
    model_input = [torch.rand(1, 1024, 14, 14) for _ in range(2)]
    model_input = MultiViewTransformerInput(features=model_input)
    output = model_intermediate_feature_returner(model_input)
    assert isinstance(output, tuple)
    assert isinstance(output[0], MultiViewTransformerOutput)
    assert len(output[1]) == 4
    assert all(isinstance(intermediate, MultiViewTransformerOutput) for intermediate in output[1])
    assert len(output[1][0].features) == 2

    # Test the normalizing of intermediate features
    model_intermediate_feature_returner = MultiViewAlternatingAttentionTransformerIFR(
        name="MV-AAT-IFR",
        input_embed_dim=1024,
        indices=[-1],  # Last layer
        norm_intermediate=False,  # Disable normalization
    )
    model_input = [torch.rand(1, 1024, 14, 14) for _ in range(2)]
    model_input = MultiViewTransformerInput(features=model_input)
    output = model_intermediate_feature_returner(model_input)
    for view_idx in range(2):
        assert not torch.equal(
            output[0].features[view_idx], output[1][-1].features[view_idx]
        ), "Final features and intermediate features (last layer) must be different."

    model_intermediate_feature_returner = MultiViewAlternatingAttentionTransformerIFR(
        name="MV-AAT-IFR",
        input_embed_dim=1024,
        indices=[-1],  # Last layer
        norm_intermediate=True,
    )
    model_input = [torch.rand(1, 1024, 14, 14) for _ in range(2)]
    model_input = MultiViewTransformerInput(features=model_input)
    output = model_intermediate_feature_returner(model_input)
    for view_idx in range(2):
        assert torch.equal(
            output[0].features[view_idx], output[1][-1].features[view_idx]
        ), "Final features and intermediate features (last layer) must be same."

    print("All Intermediate Feature Returner Tests passed!")

    # Test additonal input tokens for MultiViewAlternatingAttentionTransformer
    print("Testing MultiViewAlternatingAttentionTransformer with additional input tokens ...")
    model = MultiViewAlternatingAttentionTransformer(
        name="MV-AAT",
        input_embed_dim=1024,
    )
    num_views = 2
    num_additional_tokens = 5
    model_input = [torch.rand(1, 1024, 14, 14) for _ in range(num_views)]
    additional_tokens = torch.rand(1, 1024, num_additional_tokens)
    model_input = MultiViewTransformerInput(features=model_input, additional_input_tokens=additional_tokens)
    model_output = model(model_input)
    assert len(model_output.features) == num_views
    assert all(f.shape == (1, model.dim, 14, 14) for f in model_output.features)
    assert model_output.additional_token_features is not None
    assert model_output.additional_token_features.shape == (1, model.dim, num_additional_tokens)

    # Test additonal input tokens for MultiViewAlternatingAttentionTransformerIFR
    print("Testing MultiViewAlternatingAttentionTransformerIFR with additional input tokens ...")
    model_ifr = MultiViewAlternatingAttentionTransformerIFR(
        name="MV-AAT-IFR",
        input_embed_dim=1024,
        indices=[0, 2, 4],
    )
    model_input = [torch.rand(1, 1024, 14, 14) for _ in range(num_views)]
    additional_tokens = torch.rand(1, 1024, num_additional_tokens)
    model_input = MultiViewTransformerInput(features=model_input, additional_input_tokens=additional_tokens)
    output = model_ifr(model_input)
    assert isinstance(output, tuple)
    assert isinstance(output[0], MultiViewTransformerOutput)
    assert output[0].additional_token_features is not None
    assert output[0].additional_token_features.shape == (1, model_ifr.dim, num_additional_tokens)
    assert len(output[1]) == 3
    assert all(isinstance(intermediate, MultiViewTransformerOutput) for intermediate in output[1])
    assert all(intermediate.additional_token_features is not None for intermediate in output[1])
    assert all(
        intermediate.additional_token_features.shape == (1, model_ifr.dim, num_additional_tokens)
        for intermediate in output[1]
    )

    print("All tests using additional input tokens passed!")



"""
Adaptors for the UniCeption Prediction Heads.
"""

from functools import lru_cache
from math import isfinite
from typing import List, Tuple, Union

import numpy as np
import torch
import torch.nn as nn

from uniception.models.prediction_heads import (
    AdaptorInput,
    AdaptorOutput,
    Covariance2DAdaptorOutput,
    MaskAdaptorOutput,
    RegressionAdaptorOutput,
    RegressionWithConfidenceAdaptorOutput,
    RegressionWithConfidenceAndMaskAdaptorOutput,
    RegressionWithMaskAdaptorOutput,
    UniCeptionAdaptorBase,
)


class FlowAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        flow_mean: Union[Tuple[float, float], List[float]],
        flow_std: Union[Tuple[float, float], List[float]],
        base_shape: Tuple[int, int],
        scale_strategy: str,
        output_normalized_coordinate: bool = False,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Flow head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            flow_mean (torch.Tensor): (2,) Mean of the flow.
            flow_std (torch.Tensor): (2,) Standard deviation of the flow.
            base_shape (Tuple[int, int]): Base shape of the flow mean and std.
            scale_strategy (str): Strategy for scaling the flow, either
            - none: No scaling, network will be unnormalized with the given mean and std for all input shapes
            - scale_width: scale the output for "none" by actual width divided by base width for both X and Y
            - scale_height: scale the output for "none" by actual height divided by base height for both X and Y
            - scale_both: scale the output for "none" by actual dimension / base dimension individually for X and Y
            output_normalized_coordinate (bool): If True, will subtract the (X, Y) coordinate of the output pixel from input x after it is being scaled to pixel coordinates.
            In other words, the network will predict the pixel position that the source pixel will land on the target image, rather than the flow.
        """
        super().__init__(name, required_channels=2, *args, **kwargs)

        self.name: str = name

        flow_mean = list(flow_mean)
        flow_std = list(flow_std)

        # Handle the case where flow_mean and flow_std are passed as tuples
        if isinstance(flow_mean, tuple) or isinstance(flow_mean, list):
            flow_mean = torch.tensor(flow_mean, dtype=torch.float32)
            assert flow_mean.shape == (2,), f"Flow mean must be a 2D tensor, got {flow_mean.shape}"

        if isinstance(flow_std, tuple) or isinstance(flow_std, list):
            flow_std = torch.tensor(flow_std, dtype=torch.float32)
            assert flow_std.shape == (2,), f"Flow std must be a 2D tensor, got {flow_std.shape}"

        self.register_buffer("flow_mean", flow_mean.view(1, 2, 1, 1))
        self.register_buffer("flow_std", flow_std.view(1, 2, 1, 1))

        self.base_shape = list(base_shape)
        self.scale_strategy = scale_strategy
        self.output_normalized_coordinate = output_normalized_coordinate

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the FlowAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)

        Returns:
            AdaptorOutput: Output of the adaptor.
        """

        x = adaptor_input.adaptor_feature

        # Check the number of channels to avoid passing BHWC features
        _, C, _, _ = x.shape
        assert C == 2, f"FlowAdaptor requires BCHW format with 2 channels, got {C} channels"

        output_shape = adaptor_input.output_shape_hw

        if not self.output_normalized_coordinate:
            x_scale, y_scale = self._get_xy_scale(output_shape)

            # Scale the flow by stored mean, std and scaling factors
            flow_mean = self.flow_mean * torch.tensor([x_scale, y_scale], dtype=torch.float32, device=x.device).view(
                1, 2, 1, 1
            )
            flow_std = self.flow_std * torch.tensor([x_scale, y_scale], dtype=torch.float32, device=x.device).view(
                1, 2, 1, 1
            )

            # Unnormalize the flow
            x = x * flow_std + flow_mean
        else:
            # Optionally subtract the coordinate bias
            wh_normalizer = torch.tensor(
                adaptor_input.output_shape_hw[::-1], dtype=torch.float32, device=x.device
            ).view(1, 2, 1, 1)

            x = 0.5 * (x + 1) * wh_normalizer + 0.5

            coords = self._get_coordinate_bias(output_shape, x.device)
            x = x - coords

        return RegressionAdaptorOutput(value=x)

    def _get_xy_scale(self, output_shape: Tuple[int, int]):
        """
        Get the scaling factor for the X and Y dimensions.

        Args:
            output_shape (Tuple[int, int]): HW Shape of the output.

        Returns:
            Tuple[float, float]: Scaling factors for X and Y dimensions.
        """
        if self.scale_strategy == "none":
            return 1.0, 1.0
        elif self.scale_strategy == "scale_width":
            return output_shape[1] / self.base_shape[1], output_shape[1] / self.base_shape[1]
        elif self.scale_strategy == "scale_height":
            return output_shape[0] / self.base_shape[0], output_shape[0] / self.base_shape[0]
        elif self.scale_strategy == "scale_both":
            return output_shape[1] / self.base_shape[1], output_shape[0] / self.base_shape[0]
        else:
            raise ValueError(f"Invalid scaling strategy: {self.scale_strategy}")

    @lru_cache(maxsize=10)
    def _get_coordinate_bias(self, output_shape: Tuple[int, int], device: str):
        """
        Get the (X, Y) coordinate image for the given output shape.

        Args:
            output_shape (Tuple[int, int]): HW Shape of the output.
            device: device to store the tensor on

        Returns:
            torch.Tensor: (2, H, W) tensor with X and Y coordinates, at device. This coordinate value will
            include 0.5 px offset - i.e. the center of the top-left pixel is (0.5, 0.5).
        """

        H, W = output_shape

        coords = torch.stack(
            torch.meshgrid(
                torch.arange(0, W, device=device, dtype=torch.float32) + 0.5,
                torch.arange(0, H, device=device, dtype=torch.float32) + 0.5,
                indexing="xy",
            ),
            dim=0,
        )

        return coords


class ScaleAdaptor(UniCeptionAdaptorBase):
    def __init__(self, name: str, mode: str, vmin: float = 0, vmax: float = np.inf, *args, **kwargs):
        """
        Adaptor for scale prediction in UniCeption.

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the scale prediction, either "linear", "square" or "exp". Scales the predicted scaling factor accordingly.
            vmin (float): Minimum value of the scale prediction after scaling.
            vmax (float): Maximum value of the scale prediction after scaling.
        """
        super().__init__(name, required_channels=1, *args, **kwargs)

        self.mode = mode
        self.vmin = vmin
        self.vmax = vmax

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the ScaleAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x 1 x ...)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        predicted_scale_factor = adaptor_input.adaptor_feature
        output_scale_factor = None

        if self.mode == "linear":
            output_scale_factor = predicted_scale_factor
        elif self.mode == "square":
            output_scale_factor = predicted_scale_factor.square()
        elif self.mode == "exp":
            output_scale_factor = torch.exp(predicted_scale_factor)

        if not self.no_bounds:
            output_scale_factor = output_scale_factor.clip(self.vmin, self.vmax)

        return AdaptorOutput(value=output_scale_factor)


class DepthAdaptor(UniCeptionAdaptorBase):
    def __init__(self, name: str, mode: str, vmin: float = 0, vmax: float = np.inf, *args, **kwargs):
        """
        Adaptor for the Depth head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the depth, either "linear", "square" or "exp". Scales the depth accordingly.
            vmin (float): Minimum value of the depth after scaling.
            vmax (float): Maximum value of the depth after scaling.
        """
        super().__init__(name, required_channels=1, *args, **kwargs)

        self.mode = mode
        self.vmin = vmin
        self.vmax = vmax

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the DepthAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        x = adaptor_input.adaptor_feature
        output_depth = None

        if self.mode == "linear":
            output_depth = x
        elif self.mode == "square":
            output_depth = x**2
        elif self.mode == "exp":
            output_depth = torch.exp(x)
        else:
            raise ValueError(f"Invalid mode: {self.mode}")

        if not self.no_bounds:
            output_depth = output_depth.clip(self.vmin, self.vmax)

        return RegressionAdaptorOutput(value=output_depth)


class PointMapAdaptor(UniCeptionAdaptorBase):
    def __init__(self, name: str, mode: str, vmin: float = -np.inf, vmax: float = np.inf, *args, **kwargs):
        """
        Adaptor for the PointMap head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the point map, either "linear", "square" or "exp". Scales the distance of the points to the world origin accordingly.
            vmin (float): Minimum value of the point map after scaling.
            vmax (float): Maximum value of the point map after scaling.
        """
        super().__init__(name, required_channels=3, *args, **kwargs)

        self.mode = mode
        self.vmin = vmin
        self.vmax = vmax

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the PointMapAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        xyz = adaptor_input.adaptor_feature
        output_xyz = None

        if self.mode != "linear":
            if self.mode == "square":
                # Compute distance to world origin
                d = xyz.norm(dim=1, keepdim=True)
                output_xyz = xyz / d.clip(min=1e-8)
                # Scale the distance to world origin based on mode
                output_xyz = output_xyz * d.square()
            elif self.mode == "exp":
                # Compute distance to world origin
                d = xyz.norm(dim=1, keepdim=True)
                output_xyz = xyz / d.clip(min=1e-8)
                # Scale the distance to world origin based on mode
                output_xyz = output_xyz * torch.expm1(d)
            elif self.mode == "z_exp":
                xy, z = xyz.split([2, 1], dim=1)
                z = torch.exp(z)
                output_xyz = torch.cat([xy * z, z], dim=1)
            else:
                raise ValueError(f"Invalid mode: {self.mode}")
        else:
            output_xyz = xyz

        if not self.no_bounds:
            output_xyz = output_xyz.clip(self.vmin, self.vmax)

        return RegressionAdaptorOutput(value=output_xyz)


class RayOriginsAdaptor(UniCeptionAdaptorBase):
    def __init__(self, name: str, mode: str, vmin: float = -np.inf, vmax: float = np.inf, *args, **kwargs):
        """
        Adaptor for the RayOrigins head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the ray origins, either "linear", "square" or "exp". Scales the distance of the ray origins to the world origin accordingly.
            vmin (float): Minimum value of the ray origins after scaling.
            vmax (float): Maximum value of the ray origins after scaling.
        """
        super().__init__(name, required_channels=3, *args, **kwargs)

        self.mode = mode
        self.vmin = vmin
        self.vmax = vmax

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the RayOriginsAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        ray_origins = adaptor_input.adaptor_feature
        output_ray_origins = None

        if self.mode != "linear":
            # Compute distance to world origin
            d = ray_origins.norm(dim=1, keepdim=True)
            output_ray_origins = ray_origins / d.clip(min=1e-8)
            # Scale the distance to world origin based on mode
            if self.mode == "square":
                output_ray_origins = output_ray_origins * d.square()
            elif self.mode == "exp":
                output_ray_origins = output_ray_origins * torch.expm1(d)
            else:
                raise ValueError(f"Invalid mode: {self.mode}")
        else:
            output_ray_origins = ray_origins

        if not self.no_bounds:
            output_ray_origins = output_ray_origins.clip(self.vmin, self.vmax)

        return RegressionAdaptorOutput(value=output_ray_origins)


class RayDirectionsAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        mode: str,
        normalize_to_unit_sphere: bool,
        normalize_to_unit_image_plane: bool,
        vmin: float = -np.inf,
        vmax: float = np.inf,
        clamp_min_of_z_dir: bool = False,
        z_dir_min: float = 1,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayDirections head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the ray directions. Scales the directions accordingly. Currently only supports "linear".
            normalize_to_unit_sphere (bool): If True, will normalize the ray directions to unit vectors.
            normalize_to_unit_image_plane (bool): If True, will normalize the ray directions so that the z component is 1.
            vmin (float): Minimum value of the ray directions after scaling & before any sort of normalization. (default: -inf)
            vmax (float): Maximum value of the ray directions after scaling & before any sort of normalization. (default: inf)
            clamp_min_of_z_dir (bool): If True, will clamp the z component of the ray directions before normalization. (default: False)
            z_dir_min (float): If clamp_min_of_z_dir is True, this minimum value is used for clamping. (default: 1)
        """
        super().__init__(name, required_channels=3, *args, **kwargs)

        self.mode = mode
        self.normalize_to_unit_sphere = normalize_to_unit_sphere
        self.normalize_to_unit_image_plane = normalize_to_unit_image_plane
        self.vmin = vmin
        self.vmax = vmax
        self.clamp_min_of_z_dir = clamp_min_of_z_dir
        self.z_dir_min = z_dir_min

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the RayDirectionsAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        ray_directions = adaptor_input.adaptor_feature

        if self.mode == "linear":
            output_ray_directions = ray_directions
        else:
            raise ValueError(f"Invalid mode: {self.mode}")

        if not self.no_bounds:
            output_ray_directions = output_ray_directions.clip(self.vmin, self.vmax)

        if self.clamp_min_of_z_dir:
            # Clamp the z component of ray directions
            output_ray_directions_xy = output_ray_directions[:, :2]
            clamped_output_ray_directions_z = torch.clamp(output_ray_directions[:, 2:3], min=self.z_dir_min)
            output_ray_directions = torch.cat((output_ray_directions_xy, clamped_output_ray_directions_z), dim=1)

        if self.normalize_to_unit_sphere:
            # Normalize the ray directions to unit vectors
            output_ray_dirs_norm = output_ray_directions.norm(dim=1, keepdim=True).clip(min=1e-8)
            output_ray_directions = output_ray_directions / output_ray_dirs_norm
        elif self.normalize_to_unit_image_plane:
            # Normalize the ray directions so that the z component is 1
            output_ray_directions_z = output_ray_directions[:, 2:3]
            output_ray_directions = output_ray_directions / output_ray_directions_z

        return RegressionAdaptorOutput(value=output_ray_directions)


class RayDirectionsPlusDepthAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayDirections + Depth head in UniCeption.
        """
        super().__init__(name, required_channels=4, *args, **kwargs)

        self.ray_directions_adaptor = RayDirectionsAdaptor(
            name,
            ray_directions_mode,
            ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin,
            ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min,
        )
        self.depth_adaptor = DepthAdaptor(name, depth_mode, depth_vmin, depth_vmax)

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the RayMapPlusDepthAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        ray_directions, ray_depths = torch.split(adaptor_input.adaptor_feature, [3, 1], dim=1)
        ray_directions_adaptor_input = AdaptorInput(
            adaptor_feature=ray_directions, output_shape_hw=adaptor_input.output_shape_hw
        )
        depth_adaptor_input = AdaptorInput(adaptor_feature=ray_depths, output_shape_hw=adaptor_input.output_shape_hw)
        output_ray_directions = self.ray_directions_adaptor(ray_directions_adaptor_input)
        output_depth = self.depth_adaptor(depth_adaptor_input)
        output = torch.cat([output_ray_directions.value, output_depth.value], dim=1)

        return RegressionAdaptorOutput(value=output)


class CamTranslationAdaptor(UniCeptionAdaptorBase):
    def __init__(self, name: str, mode: str, vmin: float = -np.inf, vmax: float = np.inf, *args, **kwargs):
        """
        Adaptor for the Camera Translation or Pose head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the camera translation, either "linear", "square" or "exp". Scales the distance of the camera to the world origin accordingly.
            vmin (float): Minimum value of the camera translation after scaling.
            vmax (float): Maximum value of the camera translation after scaling.
        """
        super().__init__(name, required_channels=3, *args, **kwargs)

        self.mode = mode
        self.vmin = vmin
        self.vmax = vmax

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the CamTranslationAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C ...)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        cam_trans = adaptor_input.adaptor_feature
        output_cam_trans = None

        if self.mode != "linear":
            # Compute distance to world origin
            d = cam_trans.norm(dim=1, keepdim=True)
            output_cam_trans = cam_trans / d.clip(min=1e-8)
            # Scale the distance to world origin based on mode
            if self.mode == "square":
                output_cam_trans = output_cam_trans * d.square()
            elif self.mode == "exp":
                output_cam_trans = output_cam_trans * torch.expm1(d)
            else:
                raise ValueError(f"Invalid mode: {self.mode}")
        else:
            output_cam_trans = cam_trans

        if not self.no_bounds:
            output_cam_trans = output_cam_trans.clip(self.vmin, self.vmax)

        return AdaptorOutput(value=output_cam_trans)


class QuaternionsAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self, name: str, mode: str, normalize: bool, vmin: float = -np.inf, vmax: float = np.inf, *args, **kwargs
    ):
        """
        Adaptor for the Quaternions or Pose head in UniCeption.
        Notation of the quaternions: (x, y, z, w)

        Args:
            name (str): Name of the adaptor.
            mode (str): Mode of the quaternions. Scales the quaternions accordingly before normalization. Currently only supports "linear".
            normalize (bool): If True, will normalize the quaternions to unit quaternions.
            vmin (float): Minimum value of the quaternions after scaling & before normalization to unit quaternions if required.
            vmax (float): Maximum value of the quaternions after scaling & before normalization to unit quaternions if required.
        """
        super().__init__(name, required_channels=4, *args, **kwargs)

        self.mode = mode
        self.normalize = normalize
        self.vmin = vmin
        self.vmax = vmax

        self.no_bounds = (vmin == -float("inf")) and (vmax == float("inf"))

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the QuaternionsAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C ...)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        quaternions = adaptor_input.adaptor_feature

        if self.mode == "linear":
            output_quaternions = quaternions
        else:
            raise ValueError(f"Invalid mode: {self.mode}")

        if not self.no_bounds:
            output_quaternions = output_quaternions.clip(self.vmin, self.vmax)

        if self.normalize:
            # Normalize the quaternions to unit quaternions
            output_quats_norm = output_quaternions.norm(dim=1, keepdim=True).clip(min=1e-8)
            output_quaternions = output_quaternions / output_quats_norm

        return AdaptorOutput(value=output_quaternions)


class CamTranslationPlusQuatsAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        # Cam translation adaptor
        cam_trans_mode: str,
        cam_trans_vmin: float,
        cam_trans_vmax: float,
        # Quaternions adaptor
        quaternions_mode: str,
        quaternions_normalize: bool,
        quaternions_vmin: float,
        quaternions_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Camera Translation + Quaternions head in UniCeption.
        """
        super().__init__(name, required_channels=7, *args, **kwargs)

        self.cam_trans_adaptor = CamTranslationAdaptor(name, cam_trans_mode, cam_trans_vmin, cam_trans_vmax)
        self.quaternions_adaptor = QuaternionsAdaptor(
            name, quaternions_mode, quaternions_normalize, quaternions_vmin, quaternions_vmax
        )

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the CamTranslationPlusQuatsAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C ...)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        cam_trans, quaternions = torch.split(adaptor_input.adaptor_feature, [3, 4], dim=1)
        cam_trans_adaptor_input = AdaptorInput(adaptor_feature=cam_trans, output_shape_hw=adaptor_input.output_shape_hw)
        quaternions_adaptor_input = AdaptorInput(
            adaptor_feature=quaternions, output_shape_hw=adaptor_input.output_shape_hw
        )
        output_cam_trans = self.cam_trans_adaptor(cam_trans_adaptor_input)
        output_quaternions = self.quaternions_adaptor(quaternions_adaptor_input)
        output = torch.cat([output_cam_trans.value, output_quaternions.value], dim=1)

        return AdaptorOutput(value=output)


class RayMapAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        # Ray origins adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) head in UniCeption.
        """
        super().__init__(name, required_channels=6, *args, **kwargs)

        self.ray_origins_adaptor = RayOriginsAdaptor(name, ray_origins_mode, ray_origins_vmin, ray_origins_vmax)
        self.ray_directions_adaptor = RayDirectionsAdaptor(
            name,
            ray_directions_mode,
            ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin,
            ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min,
        )

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the RayMapAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        ray_origins, ray_directions = torch.split(adaptor_input.adaptor_feature, 3, dim=1)
        ray_origins_adaptor_input = AdaptorInput(
            adaptor_feature=ray_origins, output_shape_hw=adaptor_input.output_shape_hw
        )
        ray_directions_adaptor_input = AdaptorInput(
            adaptor_feature=ray_directions, output_shape_hw=adaptor_input.output_shape_hw
        )
        output_ray_origins = self.ray_origins_adaptor(ray_origins_adaptor_input)
        output_ray_directions = self.ray_directions_adaptor(ray_directions_adaptor_input)
        output_rays = torch.cat([output_ray_origins.value, output_ray_directions.value], dim=1)

        return RegressionAdaptorOutput(value=output_rays)


class RayMapPlusDepthAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        # Ray origins adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth head in UniCeption.
        """
        super().__init__(name, required_channels=7, *args, **kwargs)

        self.ray_origins_adaptor = RayOriginsAdaptor(name, ray_origins_mode, ray_origins_vmin, ray_origins_vmax)
        self.ray_directions_adaptor = RayDirectionsAdaptor(
            name,
            ray_directions_mode,
            ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin,
            ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min,
        )
        self.depth_adaptor = DepthAdaptor(name, depth_mode, depth_vmin, depth_vmax)

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the RayMapPlusDepthAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        ray_origins, ray_directions, ray_depths = torch.split(adaptor_input.adaptor_feature, [3, 3, 1], dim=1)
        ray_origins_adaptor_input = AdaptorInput(
            adaptor_feature=ray_origins, output_shape_hw=adaptor_input.output_shape_hw
        )
        ray_directions_adaptor_input = AdaptorInput(
            adaptor_feature=ray_directions, output_shape_hw=adaptor_input.output_shape_hw
        )
        depth_adaptor_input = AdaptorInput(adaptor_feature=ray_depths, output_shape_hw=adaptor_input.output_shape_hw)
        output_ray_origins = self.ray_origins_adaptor(ray_origins_adaptor_input)
        output_ray_directions = self.ray_directions_adaptor(ray_directions_adaptor_input)
        output_depth = self.depth_adaptor(depth_adaptor_input)
        output = torch.cat([output_ray_origins.value, output_ray_directions.value, output_depth.value], dim=1)

        return RegressionAdaptorOutput(value=output)


class RayMapPlusDepthPlusQuatsAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        # Ray origins adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Quaternions adaptor
        quaternions_mode: str,
        quaternions_normalize: bool,
        quaternions_vmin: float,
        quaternions_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth + Quaternions head in UniCeption.
        """
        super().__init__(name, required_channels=11, *args, **kwargs)

        self.ray_origins_adaptor = RayOriginsAdaptor(name, ray_origins_mode, ray_origins_vmin, ray_origins_vmax)
        self.ray_directions_adaptor = RayDirectionsAdaptor(
            name,
            ray_directions_mode,
            ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin,
            ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min,
        )
        self.depth_adaptor = DepthAdaptor(name, depth_mode, depth_vmin, depth_vmax)
        self.quaternions_adaptor = QuaternionsAdaptor(
            name, quaternions_mode, quaternions_normalize, quaternions_vmin, quaternions_vmax
        )

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the RayMapPlusDepthPlusQuatsAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """
        ray_origins, ray_directions, ray_depths, ray_quaternions = torch.split(
            adaptor_input.adaptor_feature, [3, 3, 1, 4], dim=1
        )
        ray_origins_adaptor_input = AdaptorInput(
            adaptor_feature=ray_origins, output_shape_hw=adaptor_input.output_shape_hw
        )
        ray_directions_adaptor_input = AdaptorInput(
            adaptor_feature=ray_directions, output_shape_hw=adaptor_input.output_shape_hw
        )
        depth_adaptor_input = AdaptorInput(adaptor_feature=ray_depths, output_shape_hw=adaptor_input.output_shape_hw)
        quaternions_adaptor_input = AdaptorInput(
            adaptor_feature=ray_quaternions, output_shape_hw=adaptor_input.output_shape_hw
        )
        output_ray_origins = self.ray_origins_adaptor(ray_origins_adaptor_input)
        output_ray_directions = self.ray_directions_adaptor(ray_directions_adaptor_input)
        output_ray_depths = self.depth_adaptor(depth_adaptor_input)
        output_ray_quaternions = self.quaternions_adaptor(quaternions_adaptor_input)
        output = torch.cat(
            [
                output_ray_origins.value,
                output_ray_directions.value,
                output_ray_depths.value,
                output_ray_quaternions.value,
            ],
            dim=1,
        )

        return RegressionAdaptorOutput(value=output)


class ConfidenceAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        confidence_type: str,
        vmin: float,
        vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Confidence head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            confidence_type (str): Type of the confidence, either
            - exp: Exponential confidence
            - sigmoid: Sigmoid confidence
            vmin (float): Minimum value of the confidence.
            vmax (float): Maximum value of the confidence.
        """
        super().__init__(name, required_channels=1, *args, **kwargs)

        self.confidence_type = confidence_type
        self.vmin = vmin
        self.vmax = vmax

        assert vmin < vmax, "vmin must be less than vmax"

        if confidence_type == "sigmoid":
            assert isfinite(vmin) and isfinite(vmax), "vmin and vmax must be finite for sigmoid confidence"
            assert vmin >= 0

    def forward(self, adaptor_input: AdaptorInput):
        """
        Forward pass for the ConfidenceAdaptor.

        Args:
            adaptor_input (AdaptorInput): Input to the adaptor. (B x C x H x W)
        Returns:
            AdaptorOutput: Output of the adaptor.
        """

        x = adaptor_input.adaptor_feature

        if self.confidence_type == "exp":
            confidence = self.vmin + x.exp().clip(max=self.vmax - self.vmin)

            return RegressionAdaptorOutput(value=confidence)

        elif self.confidence_type == "sigmoid":
            confidence = torch.sigmoid(x)

            confidence = confidence * (self.vmax - self.vmin) + self.vmin

            return RegressionAdaptorOutput(value=confidence)

        elif self.confidence_type == "softmax":
            B, C, H, W = x.shape
            confidence = torch.nn.functional.softmax(x.reshape(B, C, -1), dim=-1).reshape(B, C, H, W) * (H * W)

            return RegressionAdaptorOutput(value=confidence)


class Covariance2DAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        parametrization: str = "exp_tanh",
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Covariance2D head in UniCeption.
        """
        super().__init__(name, required_channels=3, *args, **kwargs)
        self.parametrization = parametrization

    def forward(self, adaptor_input: AdaptorInput):
        x = adaptor_input.adaptor_feature

        if self.parametrization == "exp_tanh":
            c1, c2, s = torch.split(x, 1, dim=1)

            diag_exponent = (c1 + c2) / 2
            tanh_s = s.tanh()

            cov = torch.cat([c1.exp(), c2.exp(), tanh_s * torch.exp(diag_exponent)], dim=1)

            log_det = c1 + c2 + torch.log(1 - torch.square(tanh_s) + 1e-8)

            inv_coeff = 1 / (1 - torch.square(tanh_s) + 1e-8)
            inv_cov = inv_coeff * torch.cat(
                [torch.exp(-c1), torch.exp(-c2), -tanh_s * torch.exp(-diag_exponent)], dim=1
            )

        else:
            raise ValueError(f"Invalid parametrization: {self.parametrization}")

        return Covariance2DAdaptorOutput(covariance=cov, log_det=log_det, inv_covariance=inv_cov)


class MaskAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Mask head in UniCeption.
        """
        super().__init__(name, required_channels=1, *args, **kwargs)

    def forward(self, adaptor_input: AdaptorInput):
        x = adaptor_input.adaptor_feature

        mask = torch.sigmoid(x)

        return MaskAdaptorOutput(logits=x, mask=mask)


class ValueWithConfidenceAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        value_adaptor: UniCeptionAdaptorBase,
        confidence_adaptor: UniCeptionAdaptorBase,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Value with Confidence head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            value_adaptor (UniCeptionAdaptorBase): Adaptor for the value.
            confidence_adaptor (UniCeptionAdaptorBase): Adaptor for the confidence.
        """

        super().__init__(
            name,
            required_channels=value_adaptor.required_channels + confidence_adaptor.required_channels,
            *args,
            **kwargs,
        )

        self.value_adaptor = value_adaptor
        self.confidence_adaptor = confidence_adaptor

    def forward(self, adaptor_input: AdaptorInput):
        value_input, confidence_input = torch.split(
            adaptor_input.adaptor_feature,
            [self.value_adaptor.required_channels, self.confidence_adaptor.required_channels],
            dim=1,
        )
        value_adaptor_input = AdaptorInput(adaptor_feature=value_input, output_shape_hw=adaptor_input.output_shape_hw)
        confidence_adaptor_input = AdaptorInput(
            adaptor_feature=confidence_input, output_shape_hw=adaptor_input.output_shape_hw
        )
        value_output = self.value_adaptor(value_adaptor_input)
        confidence_output = self.confidence_adaptor(confidence_adaptor_input)

        return RegressionWithConfidenceAdaptorOutput(value=value_output.value, confidence=confidence_output.value)


class FlowWithConfidenceAdaptor(ValueWithConfidenceAdaptor):
    def __init__(
        self,
        name: str,
        # Flow adaptor
        flow_mean: torch.Tensor,
        flow_std: torch.Tensor,
        base_shape: Tuple[int, int],
        scale_strategy: str,
        output_normalized_coordinate: bool,
        # Confidence adaptor
        confidence_type: str,
        vmin: float,
        vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Flow with Confidence head in UniCeption.
        """
        flow_adaptor = FlowAdaptor(
            name=f"{name}",
            flow_mean=flow_mean,
            flow_std=flow_std,
            base_shape=base_shape,
            scale_strategy=scale_strategy,
            output_normalized_coordinate=output_normalized_coordinate,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=vmin, vmax=vmax
        )

        super().__init__(name, value_adaptor=flow_adaptor, confidence_adaptor=confidence_adaptor, *args, **kwargs)


class PointMapWithConfidenceAdaptor(ValueWithConfidenceAdaptor):
    def __init__(
        self,
        name: str,
        # Pointmap adaptor
        pointmap_mode: str,
        pointmap_vmin: float,
        pointmap_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the PointMap with Confidence head in UniCeption.
        """
        pointmap_adaptor = PointMapAdaptor(name=f"{name}", mode=pointmap_mode, vmin=pointmap_vmin, vmax=pointmap_vmax)

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        super().__init__(name, value_adaptor=pointmap_adaptor, confidence_adaptor=confidence_adaptor, *args, **kwargs)


class RayDirectionsPlusDepthwithConfidenceAdaptor(ValueWithConfidenceAdaptor):
    def __init__(
        self,
        name: str,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayDirections + Depth with Confidence head in UniCeption.
        """
        ray_directions_plus_depth_adaptor = RayDirectionsPlusDepthAdaptor(
            name=f"{name}",
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        super().__init__(
            name,
            value_adaptor=ray_directions_plus_depth_adaptor,
            confidence_adaptor=confidence_adaptor,
            *args,
            **kwargs,
        )


class RayMapPlusDepthwithConfidenceAdaptor(ValueWithConfidenceAdaptor):
    def __init__(
        self,
        name: str,
        # RayMap adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth with Confidence head in UniCeption.
        """
        raymap_plus_depth_adaptor = RayMapPlusDepthAdaptor(
            name=f"{name}",
            ray_origins_mode=ray_origins_mode,
            ray_origins_vmin=ray_origins_vmin,
            ray_origins_vmax=ray_origins_vmax,
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        super().__init__(
            name, value_adaptor=raymap_plus_depth_adaptor, confidence_adaptor=confidence_adaptor, *args, **kwargs
        )


class RayMapPlusDepthPlusQuatswithConfidenceAdaptor(ValueWithConfidenceAdaptor):
    def __init__(
        self,
        name: str,
        # RayMap adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Quaternions adaptor
        quaternions_mode: str,
        quaternions_normalize: bool,
        quaternions_vmin: float,
        quaternions_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth + Quaternions with Confidence head in UniCeption.
        """
        raymap_plus_depth_plus_quats_adaptor = RayMapPlusDepthPlusQuatsAdaptor(
            name=f"{name}",
            ray_origins_mode=ray_origins_mode,
            ray_origins_vmin=ray_origins_vmin,
            ray_origins_vmax=ray_origins_vmax,
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
            quaternions_mode=quaternions_mode,
            quaternions_normalize=quaternions_normalize,
            quaternions_vmin=quaternions_vmin,
            quaternions_vmax=quaternions_vmax,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        super().__init__(
            name,
            value_adaptor=raymap_plus_depth_plus_quats_adaptor,
            confidence_adaptor=confidence_adaptor,
            *args,
            **kwargs,
        )


class ValueWithMaskAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        value_adaptor: UniCeptionAdaptorBase,
        mask_adaptor: UniCeptionAdaptorBase,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Value with Mask head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            value_adaptor (UniCeptionAdaptorBase): Adaptor for the value.
            mask_adaptor (UniCeptionAdaptorBase): Adaptor for the mask.
        """

        super().__init__(
            name,
            required_channels=value_adaptor.required_channels + mask_adaptor.required_channels,
            *args,
            **kwargs,
        )

        self.value_adaptor = value_adaptor
        self.mask_adaptor = mask_adaptor

    def forward(self, adaptor_input: AdaptorInput):
        value_input, mask_input = torch.split(
            adaptor_input.adaptor_feature,
            [self.value_adaptor.required_channels, self.mask_adaptor.required_channels],
            dim=1,
        )
        value_adaptor_input = AdaptorInput(adaptor_feature=value_input, output_shape_hw=adaptor_input.output_shape_hw)
        mask_adaptor_input = AdaptorInput(adaptor_feature=mask_input, output_shape_hw=adaptor_input.output_shape_hw)
        value_output = self.value_adaptor(value_adaptor_input)
        mask_output = self.mask_adaptor(mask_adaptor_input)

        return RegressionWithMaskAdaptorOutput(
            value=value_output.value, mask=mask_output.mask, logits=mask_output.logits
        )


class PointMapWithMaskAdaptor(ValueWithMaskAdaptor):
    def __init__(
        self,
        name: str,
        # Pointmap adaptor
        pointmap_mode: str,
        pointmap_vmin: float,
        pointmap_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the PointMap with Confidence head in UniCeption.
        """
        pointmap_adaptor = PointMapAdaptor(name=f"{name}", mode=pointmap_mode, vmin=pointmap_vmin, vmax=pointmap_vmax)

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(name, value_adaptor=pointmap_adaptor, mask_adaptor=mask_adaptor, *args, **kwargs)


class RayDirectionsPlusDepthwithMaskAdaptor(ValueWithMaskAdaptor):
    def __init__(
        self,
        name: str,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayDirections + Depth with Mask head in UniCeption.
        """
        ray_directions_plus_depth_adaptor = RayDirectionsPlusDepthAdaptor(
            name=f"{name}",
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(
            name, value_adaptor=ray_directions_plus_depth_adaptor, mask_adaptor=mask_adaptor, *args, **kwargs
        )


class RayMapPlusDepthwithMaskAdaptor(ValueWithMaskAdaptor):
    def __init__(
        self,
        name: str,
        # RayMap adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth with Mask head in UniCeption.
        """
        raymap_plus_depth_adaptor = RayMapPlusDepthAdaptor(
            name=f"{name}",
            ray_origins_mode=ray_origins_mode,
            ray_origins_vmin=ray_origins_vmin,
            ray_origins_vmax=ray_origins_vmax,
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(name, value_adaptor=raymap_plus_depth_adaptor, mask_adaptor=mask_adaptor, *args, **kwargs)


class RayMapPlusDepthPlusQuatswithMaskAdaptor(ValueWithMaskAdaptor):
    def __init__(
        self,
        name: str,
        # RayMap adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Quaternions adaptor
        quaternions_mode: str,
        quaternions_normalize: bool,
        quaternions_vmin: float,
        quaternions_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth + Quaternions with Mask head in UniCeption.
        """
        raymap_plus_depth_plus_quats_adaptor = RayMapPlusDepthPlusQuatsAdaptor(
            name=f"{name}",
            ray_origins_mode=ray_origins_mode,
            ray_origins_vmin=ray_origins_vmin,
            ray_origins_vmax=ray_origins_vmax,
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
            quaternions_mode=quaternions_mode,
            quaternions_normalize=quaternions_normalize,
            quaternions_vmin=quaternions_vmin,
            quaternions_vmax=quaternions_vmax,
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(
            name, value_adaptor=raymap_plus_depth_plus_quats_adaptor, mask_adaptor=mask_adaptor, *args, **kwargs
        )


class ValueWithConfidenceAndMaskAdaptor(UniCeptionAdaptorBase):
    def __init__(
        self,
        name: str,
        value_adaptor: UniCeptionAdaptorBase,
        confidence_adaptor: UniCeptionAdaptorBase,
        mask_adaptor: UniCeptionAdaptorBase,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the Value with Confidence & Mask head in UniCeption.

        Args:
            name (str): Name of the adaptor.
            value_adaptor (UniCeptionAdaptorBase): Adaptor for the value.
            mask_adaptor (UniCeptionAdaptorBase): Adaptor for the mask.
        """

        super().__init__(
            name,
            required_channels=value_adaptor.required_channels
            + confidence_adaptor.required_channels
            + mask_adaptor.required_channels,
            *args,
            **kwargs,
        )

        self.value_adaptor = value_adaptor
        self.confidence_adaptor = confidence_adaptor
        self.mask_adaptor = mask_adaptor

    def forward(self, adaptor_input: AdaptorInput):
        value_input, confidence_input, mask_input = torch.split(
            adaptor_input.adaptor_feature,
            [
                self.value_adaptor.required_channels,
                self.confidence_adaptor.required_channels,
                self.mask_adaptor.required_channels,
            ],
            dim=1,
        )
        value_adaptor_input = AdaptorInput(adaptor_feature=value_input, output_shape_hw=adaptor_input.output_shape_hw)
        confidence_adaptor_input = AdaptorInput(
            adaptor_feature=confidence_input, output_shape_hw=adaptor_input.output_shape_hw
        )
        mask_adaptor_input = AdaptorInput(adaptor_feature=mask_input, output_shape_hw=adaptor_input.output_shape_hw)
        value_output = self.value_adaptor(value_adaptor_input)
        confidence_output = self.confidence_adaptor(confidence_adaptor_input)
        mask_output = self.mask_adaptor(mask_adaptor_input)

        return RegressionWithConfidenceAndMaskAdaptorOutput(
            value=value_output.value,
            confidence=confidence_output.value,
            mask=mask_output.mask,
            logits=mask_output.logits,
        )


class PointMapWithConfidenceAndMaskAdaptor(ValueWithConfidenceAndMaskAdaptor):
    def __init__(
        self,
        name: str,
        # PointMap adaptor
        pointmap_mode: str,
        pointmap_vmin: float,
        pointmap_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the PointMap with Confidence & Mask head in UniCeption.
        """
        pointmap_adaptor = PointMapAdaptor(name=f"{name}", mode=pointmap_mode, vmin=pointmap_vmin, vmax=pointmap_vmax)

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(
            name,
            value_adaptor=pointmap_adaptor,
            confidence_adaptor=confidence_adaptor,
            mask_adaptor=mask_adaptor,
            *args,
            **kwargs,
        )


class RayDirectionsPlusDepthwithConfidenceAndMaskAdaptor(ValueWithConfidenceAndMaskAdaptor):
    def __init__(
        self,
        name: str,
        # Ray directions adaptor
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayDirections + Depth with Confidence & Mask head in UniCeption.
        """
        ray_directions_plus_depth_adaptor = RayDirectionsPlusDepthAdaptor(
            name=f"{name}",
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(
            name,
            value_adaptor=ray_directions_plus_depth_adaptor,
            confidence_adaptor=confidence_adaptor,
            mask_adaptor=mask_adaptor,
            *args,
            **kwargs,
        )


class RayMapPlusDepthwithConfidenceAndMaskAdaptor(ValueWithConfidenceAndMaskAdaptor):
    def __init__(
        self,
        name: str,
        # RayMap adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth with Confidence & Mask head in UniCeption.
        """
        raymap_plus_depth_adaptor = RayMapPlusDepthAdaptor(
            name=f"{name}",
            ray_origins_mode=ray_origins_mode,
            ray_origins_vmin=ray_origins_vmin,
            ray_origins_vmax=ray_origins_vmax,
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(
            name,
            value_adaptor=raymap_plus_depth_adaptor,
            confidence_adaptor=confidence_adaptor,
            mask_adaptor=mask_adaptor,
            *args,
            **kwargs,
        )


class RayMapPlusDepthPlusQuatswithConfidenceAndMaskAdaptor(ValueWithConfidenceAndMaskAdaptor):
    def __init__(
        self,
        name: str,
        # RayMap adaptor
        ray_origins_mode: str,
        ray_origins_vmin: float,
        ray_origins_vmax: float,
        ray_directions_mode: str,
        ray_directions_normalize_to_unit_sphere: bool,
        ray_directions_normalize_to_unit_image_plane: bool,
        ray_directions_vmin: float,
        ray_directions_vmax: float,
        ray_directions_clamp_min_of_z_dir: bool,
        ray_directions_z_dir_min: float,
        # Depth adaptor
        depth_mode: str,
        depth_vmin: float,
        depth_vmax: float,
        # Quaternions adaptor
        quaternions_mode: str,
        quaternions_normalize: bool,
        quaternions_vmin: float,
        quaternions_vmax: float,
        # Confidence adaptor
        confidence_type: str,
        confidence_vmin: float,
        confidence_vmax: float,
        *args,
        **kwargs,
    ):
        """
        Adaptor for the RayMap (RayOrigins + RayDirections) + Depth + Quaternions with Confidence & Mask head in UniCeption.
        """
        raymap_plus_depth_plus_quats_adaptor = RayMapPlusDepthPlusQuatsAdaptor(
            name=f"{name}",
            ray_origins_mode=ray_origins_mode,
            ray_origins_vmin=ray_origins_vmin,
            ray_origins_vmax=ray_origins_vmax,
            ray_directions_mode=ray_directions_mode,
            ray_directions_normalize_to_unit_sphere=ray_directions_normalize_to_unit_sphere,
            ray_directions_normalize_to_unit_image_plane=ray_directions_normalize_to_unit_image_plane,
            ray_directions_vmin=ray_directions_vmin,
            ray_directions_vmax=ray_directions_vmax,
            ray_directions_clamp_min_of_z_dir=ray_directions_clamp_min_of_z_dir,
            ray_directions_z_dir_min=ray_directions_z_dir_min,
            depth_mode=depth_mode,
            depth_vmin=depth_vmin,
            depth_vmax=depth_vmax,
            quaternions_mode=quaternions_mode,
            quaternions_normalize=quaternions_normalize,
            quaternions_vmin=quaternions_vmin,
            quaternions_vmax=quaternions_vmax,
        )

        confidence_adaptor = ConfidenceAdaptor(
            name=f"{name}_confidence", confidence_type=confidence_type, vmin=confidence_vmin, vmax=confidence_vmax
        )

        mask_adaptor = MaskAdaptor(name=f"{name}_mask")

        super().__init__(
            name,
            value_adaptor=raymap_plus_depth_plus_quats_adaptor,
            confidence_adaptor=confidence_adaptor,
            mask_adaptor=mask_adaptor,
            *args,
            **kwargs,
        )
